import argparse
from datetime import datetime, timedelta
from enum import Enum
import json
import logging
import sys
from time import time

import numpy as np

from pyschism import dates
from pyschism.enums import Stratification, Sflux1Types, Sflux2Types
from pyschism.mesh import Hgrid, Vgrid, gridgr3
from pyschism.forcing import nws, bctides, hycom, source_sink
from pyschism.utils.timedeltatype import TimeDeltaType


logger = logging.getLogger(__name__)


def add_log_level_to_parser(parser):
    parser.add_argument(
        "--log-level",
        choices=["warning", "info", "debug"],
        default="warning",
    )


class HgridAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if self.const is None and not bool(
            set(sys.argv).intersection(["-h", "--help"])
        ):
            tmp_parser = argparse.ArgumentParser(add_help=False)
            tmp_parser.add_argument("--hgrid-crs")
            tmp_args, _ = tmp_parser.parse_known_args()
            logger.info(f"Opening hgrid from {values}...")
            start = time()
            hgrid = Hgrid.open(values, crs=tmp_args.hgrid_crs)
            logger.info(f"Reading hgrid took {time()-start}...")
            setattr(namespace, self.dest, hgrid)
        else:
            setattr(namespace, self.dest, self.const)


def add_hgrid_to_parser(parser, const=None):
    hgrid = parser.add_argument_group("Horizontal grid options")
    hgrid.add_argument(
        "hgrid",
        action=HgridAction,  # ,
        const=const,
        help="Path to the SCHISM hgrid file.",
    )
    hgrid.add_argument(
        "--hgrid-crs", help="Coordinate reference system string of hgrid."
    )


class VgridAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if self.const is None and not bool(
            set(sys.argv).intersection(["-h", "--help"])
        ):  # This is the init
            logger.info(f"Opening vgrid from {values}")
            start = time()
            vgrid = Vgrid.open(values)
            logger.info(f"Opening vgrid took {time()-start}")
            setattr(namespace, self.dest, vgrid)
        else:
            setattr(namespace, self.dest, self.const)


class VgridBinAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, Vgrid.from_binary(namespace.hgrid, values))


def add_vgrid_to_parser(parser, const=None):
    vgrid_group = parser.add_argument_group("Vertical grid options")
    vgrid = vgrid_group.add_mutually_exclusive_group()
    vgrid.add_argument(
        "--vgrid",
        action=VgridAction,  # if not bool(set(sys.argv).intersection(['-h', '--help'])) else None,
        default=Vgrid.default(),
        const=const,
        metavar="PATH_TO_FILE",
        help="Path to vgrid.in file to use in model.",
    )
    vgrid.add_argument(
        "--vgrid-bin",
        action=VgridBinAction,
        nargs="?",
        metavar="PATH_TO_BINARY",
        help="Path to the vgrid generation binary (must exist in PATH). "
        "Defaults to gen_vqs.",
    )


def add_fgrid_to_parser(parser):
    fgrid_group = parser.add_argument_group("Friction grid options.")
    fgrid_group.add_argument("--fgrid", help="Friction grid file.")
    fgrid_group.add_argument("--fgrid-crs")
    fgrid_group.add_argument(
        "--fgrid-type",
        choices=["auto", "manning", "drag", "rough"],
        default="auto",
        help="Can be used to specify which friction type the file is.",
    )


def add_stratification_to_parser(parser):

    stratification_group = parser.add_argument_group(
        "Stratification options."
    ).add_mutually_exclusive_group()
    stratification_group.add_argument(
        "--baroclinic",
        dest="stratification",
        action="store_const",
        const=Stratification.BAROCLINIC,
    )
    stratification_group.add_argument(
        "--barotropic",
        dest="stratification",
        action="store_const",
        const=Stratification.BAROTROPIC,
    )

    # args = parser.parse_known_args()[0]
    # stratification = (
    #     Stratification.BAROTROPIC
    #     if args.vgrid.is2D() is True
    #     else Stratification.BAROCLINIC
    # )
    # stratification_group.set_default(stratification=None)


def add_albedo_to_parser(parser):
    parser.add_argument("--albedo")


def add_diffmin_to_parser(parser):
    parser.add_argument("--diffmin")


def add_diffmax_to_parser(parser):
    parser.add_argument("--diffmax")


def add_watertype_to_parser(parser):
    parser.add_argument("--watertype")


def add_gridgr3_to_parser(parser):
    gridgr3_group = parser.add_argument_group("Auxiliary fields.")
    add_albedo_to_parser(gridgr3_group)
    add_diffmin_to_parser(gridgr3_group)
    add_diffmax_to_parser(gridgr3_group)
    add_watertype_to_parser(gridgr3_group)


def add_tvdflag_to_parser(parser):
    parser.add_argument("--tvdflag")


def add_fluxflag_to_parser(parser):
    parser.add_argument("--fluxflag")


def add_prop_to_parser(parser):
    prop_group = parser.add_argument_group("Prop fields.")
    add_tvdflag_to_parser(prop_group)
    add_fluxflag_to_parser(prop_group)


def add_tidal_constituents_to_parser(parser):

    # TODO: Get list directly from objects
    choices = ["K1", "O1", "P1", "Q1", "MM", "Mf", "M4", "MN4", "MS4", "2N2", "S1"]

    # class AllConstituentsAction(argparse.Action):
    #     def __call__(self, parser, namespace, values, option_string=None):
    #         tmp_parser = argparse.ArgumentParser(add_help=False)
    #         add_tidal_database_to_parser(tmp_parser)
    #         tmp_args = tmp_parser.parse_known_args()[0]
    #         tides = bctides.Tides(
    #             tidal_database=tmp_args.tidal_database,
    #             elevation=False,
    #             velocity=False,
    #         )
    #         tides.use_all()
    #         setattr(namespace, self.dest, tides)

    tide_group = parser.add_argument_group(
        "Tidal constituent options"
    ).add_mutually_exclusive_group()
    tide_group.add_argument(
        "--all-constituents",
        action="store_const",
        const="all",
        dest="constituents",
        help="Enables the use of all available tidal constituents. (default)",
    )
    tide_group.add_argument(
        "--major-constituents",
        action="store_const",
        const="major",
        dest="constituents",
        help="Enables the use of major group  of constituents only.",
    )
    tide_group.add_argument(
        "-c",
        "--constituents",
        action="append",
        choices=choices,
        dest="constituents",
        help="Enables specific tidal constituent to be forced in the model. "
        "(case-insensitive)",
    )
    tide_group.set_defaults(constituents="all")
    # tmp_parser = argparse.ArgumentParser(add_help=False)
    # add_tidal_database_to_parser(tmp_parser)
    # tmp_args = tmp_parser.parse_known_args()[0]
    # tides = bctides.Tides(
    #     tidal_database=tmp_args.tidal_database,
    #     elevation=False,
    #     velocity=False,
    # )
    # tides.use_all()
    # tide_group.set_defaults(tides=tides)


def add_bctides_options_to_parser(parser):
    bctides_options_group = parser.add_argument_group("Bctides options")
    bctides_options_group.add_argument("--cutoff-depth", type=float, default=50.0)
    bctides_options_group.add_argument(
        "--Z0",
        type=float,
    )


def add_tidal_database_to_parser(parser):

    # class TidalDatabaseAction(argparse.Action):
    #     def __call__(self, parser, namespace, values, option_string=None):
    #         tmp_parser = argparse.ArgumentParser(add_help=False)
    #         add_tidal_constituents_to_parser(tmp_parser)
    #         tmp_args = tmp_parser.parse_known_args()[0]
    #         print(tmp_args)
    #         exit()

    tidal_db_group = parser.add_argument_group("Tidal harmonics database options")
    tidal_db_group.add_argument(
        "--tidal-database",
        choices=["tpxo", "hamtide"],
        default="tpxo",
        # action=TidalDatabaseAction,
        help="Selects which tidal harmonics database to use. (default=tpxo)",
    )


# def add_baroclinic_database_to_parser(parser):
#     class BaroclinicDatabases(Enum):
#         RTOFS = hycom.RTOFS
#         GOFS = hycom.GOFS
# 
#     class BaroclinicDatabaseAction(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
# 
#             setattr(namespace, self.dest, values)
# 
#     hycom_group = parser.add_argument_group("HYCOM options")
#     # *************************************************************************************************
#     hycom_group.add_argument(
#         "--hycom",
#         # "--baroclinic-database",
#         choices=[x.name.lower() for x in BaroclinicDatabases],
#         dest="baroclinic_database",
#         default="gofs",
#         action=BaroclinicDatabaseAction,
#         help="Options for obtaining HYCOM data. Defaults to GOFS.",
#     )


# class CustomBoundaryAction(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
#         if isinstance(values, str):
#             if "iettype" in option_string:
#                 values = json.loads(values)
#                 for bnd_id, tval in values.items():
#                     if not isinstance(tval, int):
#                         raise TypeError(
#                             'Boundary type must be an integer, not type '
#                             f'{type(tval)}.')
#                     if tval == 3:
#                         namespace.hgrid.boundaries.set_forcing(
#                             bnd_id,
#                             iettype=iettype.Iettype3(get_tides(namespace))
#                             )
#                     else:
#                         raise ValueError
#         setattr(namespace, self.dest, values)


def get_per_boundary_help_message(varname, ibctype):
    return (
        f"Per-boundary specification option for {varname} variable. "
        "The format required is a json-style string. For example: "
        f"--{ibctype}=" + "'{\"1\": 3}' "
        f" would apply --{ibctype}-3 to boundary "
        'with id equal to "1".'
    )


# def add_iettype_to_parser(parser):
#     class IettypeAction(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             values = json.loads(values)
#             for key, val in values.items():
#                 values.update({key: self.iettype(int(key), val)})
#             setattr(namespace, self.dest, values)
# 
#         def iettype(self, key, value):
#             values = [1, 2, 3, 4, 5, -1]
#             assert value in values, (
#                 "per-boundary iettype values must be " f"one of {values}, not {value}."
#             )
#             raise NotImplementedError(
#                 "Per-boundary specification not yet " "supported."
#             )
# 
#     class Iettype3Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             tmp_parser = argparse.ArgumentParser(add_help=False)
#             add_tidal_constituents_to_parser(tmp_parser)
#             add_tidal_database_to_parser(tmp_parser)
#             tmp_args = tmp_parser.parse_known_args()[0]
#             setattr(
#                 namespace,
#                 self.dest,
#                 self.const(
#                     constituents=tmp_args.constituents, database=tmp_args.tidal_database
#                 ),
#             )
# 
#     class Iettype4Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             if not bool(set(sys.argv).intersection(["-h", "--help"])):
#                 if namespace.vgrid.is2D() is True:
#                     raise NotImplementedError("--iettype-4 not available for 2D model.")
#                 tmp_parser = argparse.ArgumentParser(add_help=False)
#                 add_baroclinic_database_to_parser(tmp_parser)
#                 tmp_args = tmp_parser.parse_known_args()[0]
#                 setattr(namespace, self.dest, self.const(tmp_args.baroclinic_database))
# 
#     class Iettype5Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             if not bool(set(sys.argv).intersection(["-h", "--help"])):
#                 if namespace.vgrid.is2D() is True:
#                     raise NotImplementedError("--iettype-5 not available for 2D model.")
#                 tmp_parser = argparse.ArgumentParser(add_help=False)
#                 add_tidal_constituents_to_parser(tmp_parser)
#                 add_tidal_database_to_parser(tmp_parser)
#                 add_baroclinic_database_to_parser(tmp_parser)
#                 tmp_args = tmp_parser.parse_known_args()[0]
#                 iettype3 = bctides.iettype.Iettype3(
#                     constituents=tmp_args.constituents, database=tmp_args.tidal_database
#                 )
#                 iettype4 = bctides.iettype.Iettype4(
#                     data_source=tmp_args.baroclinic_database
#                 )
#                 setattr(namespace, self.dest, self.const(iettype3, iettype4))
# 
#     iettype_group = parser.add_argument_group(
#         "Elevation boundary condition options"
#     ).add_mutually_exclusive_group()
#     iettype_group.add_argument(
#         "--iettype",
#         "--per-boundary-elevation",
#         dest="iettype",
#         metavar="JSON_STYLE_STRING",
#         action=IettypeAction,
#         # const=bctides.iettype.Iettype,
#         help=get_per_boundary_help_message("elevation", "iettype"),
#     )
#     iettype_group.add_argument(
#         "--iettype-1",
#         "--elevation-time-history",
#         dest="iettype",
#         metavar="ELEVATION_TIME_HISTORY",
#         type=bctides.iettype.Iettype1,
#         help="Global elevation options for time history.  (Not implemented)",
#     )
#     iettype_group.add_argument(
#         "--iettype-2",
#         "--constant-elevation-value",
#         dest="iettype",
#         metavar="VALUE",
#         type=bctides.iettype.Iettype2,
#         help="Global constant elevation option.",
#     )
# 
#     iettype_group.add_argument(
#         "--iettype-3",
#         "--harmonic-tidal-elevation",
#         dest="iettype",
#         nargs=0,
#         const=bctides.iettype.Iettype3,
#         action=Iettype3Action,
#         help="Enables tidal elevation harmonic forcing.",
#     )
# 
#     iettype_group.add_argument(
#         "--iettype-4",
#         "--subtidal-elevation",
#         "--elev-2d",
#         dest="iettype",
#         nargs=0,
#         const=bctides.iettype.Iettype4,
#         action=Iettype4Action,
#         help="Enables subtidal elevation forcing.",
#     )
# 
#     iettype_group.add_argument(
#         "--iettype-5",
#         "--subtidal-and-tidal-elevation",
#         dest="iettype",
#         nargs=0,
#         const=bctides.iettype.Iettype5,
#         action=Iettype5Action,
#         help="Enables the combined elevation tidal harmonic forcing and "
#         "subtidal elevation forcing.",
#     )
# 
#     iettype_group.add_argument(
#         "--iettype--1",
#         "--zero-elevation",
#         dest="iettype",
#         const=bctides.iettype.Iettype_1,
#         action="store_const",
#         help="Zero elevation at boundaries. (Not implemented)",
#     )
# 
# 
# def add_ifltype_to_parser(parser):
#     class IfltypeAction(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             values = json.loads(values)
#             for key, val in values.items():
#                 values.update({key: self.iettype(int(key), val)})
#             setattr(namespace, self.dest, values)
# 
#         def ifltype(self, key, value):
#             values = [1, 2, 3, 4, 5, -1]
#             assert value in values, (
#                 "per-boundary ifltype values must be " f"one of {values}, not {value}."
#             )
#             raise NotImplementedError(
#                 "Per-boundary specification not yet " "supported."
#             )
# 
#     class Ifltype3Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             tmp_parser = argparse.ArgumentParser(add_help=False)
#             add_tidal_constituents_to_parser(tmp_parser)
#             add_tidal_database_to_parser(tmp_parser)
#             tmp_args = tmp_parser.parse_known_args()[0]
#             setattr(
#                 namespace,
#                 self.dest,
#                 self.const(
#                     constituents=tmp_args.constituents,
#                     database=tmp_args.tidal_database,
#                 ),
#             )
# 
#     class Ifltype4Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             if not bool(set(sys.argv).intersection(["-h", "--help"])):
#                 if namespace.vgrid.is2D() is True:
#                     raise NotImplementedError("--ifltype-4 not available for 2D model.")
#                 tmp_parser = argparse.ArgumentParser(add_help=False)
#                 add_baroclinic_database_to_parser(tmp_parser)
#                 tmp_args = tmp_parser.parse_known_args()[0]
#                 setattr(namespace, self.dest, self.const(tmp_args.baroclinic_database))
# 
#     class Ifltype5Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             if not bool(set(sys.argv).intersection(["-h", "--help"])):
#                 if namespace.vgrid.is2D() is True:
#                     raise NotImplementedError("--ifltype-5 not available for 2D model.")
#                 tmp_parser = argparse.ArgumentParser(add_help=False)
#                 add_tidal_constituents_to_parser(tmp_parser)
#                 add_tidal_database_to_parser(tmp_parser)
#                 add_baroclinic_database_to_parser(tmp_parser)
#                 tmp_args = tmp_parser.parse_known_args()[0]
#                 ifltype3 = bctides.ifltype.Ifltype3(
#                     constituents=tmp_args.constituents, database=tmp_args.tidal_database
#                 )
#                 ifltype4 = bctides.ifltype.Ifltype4(
#                     data_source=tmp_args.baroclinic_database
#                 )
#                 setattr(namespace, self.dest, self.const(ifltype3, ifltype4))
# 
#     ifltype_group = parser.add_argument_group(
#         "Velocity boundary condition options"
#     ).add_mutually_exclusive_group()
#     # ifltype_group.add_argument(
#     #     "--ifltype",
#     #     '--per-boundary-velocity',
#     #     dest="ifltype",
#     #     metavar='JSON_STYLE_STRING',
#     #     action=IfltypeAction,
#     #     help=get_per_boundary_help_message('velocity', 'ifltype')
#     # )
#     ifltype_group.add_argument(
#         "--ifltype-1",
#         "--flux-time-history",
#         metavar="VELOCITY_TIME_HISTORY",
#         dest="ifltype",
#         # type=ifltype.Ifltype1
#         help="Global boundary velocity time history file. (Not implemented)",
#     )
#     ifltype_group.add_argument(
#         "--ifltype-2",
#         "--constant-flux-value",
#         dest="ifltype",
#         metavar="VALUE",
#         # type=ifltype.Ifltype2,
#         help="Global constant flow option.",
#     )
#     ifltype_group.add_argument(
#         "--ifltype-3",
#         "--harmonic-tidal-velocity",
#         dest="ifltype",
#         nargs=0,
#         const=bctides.ifltype.Ifltype3,
#         action=Ifltype3Action,
#         help="Enables tidal velocity harmonic forcing.",
#     )
# 
#     ifltype_group.add_argument(
#         "--ifltype-4",
#         "--subtidal-velocity",
#         dest="ifltype",
#         nargs=0,
#         const=bctides.ifltype.Ifltype4,
#         action=Ifltype4Action,
#         help="Enables subtidal velocity forcing.",
#     )
# 
#     ifltype_group.add_argument(
#         "--ifltype-5",
#         "--subtidal-and-tidal-velocity",
#         action=Ifltype5Action,
#         nargs=0,
#         dest="ifltype",
#         const=bctides.ifltype.Ifltype5,
#         help="Enables the combined velocity tidal harmonic forcing and "
#         "subtidal velocity forcing.",
#     )
# 
#     ifltype_group.add_argument(
#         "--ifltype--1",
#         "--uv-zero",
#         "--flather",
#         # action='store_const',
#         dest="ifltype",
#         # const=ifltype.Ifltype_1,
#     )
# 
# 
# def add_itetype_to_parser(parser):
#     def get_nudge_bnds(namespace):
#         tmp_parser = argparse.ArgumentParser(add_help=False)
#         add_baroclinic_database_to_parser(tmp_parser)
#         add_nudge_to_parser(tmp_parser)
#         tmp_args = tmp_parser.parse_known_args()[0]
#         bnd_ids = namespace.hgrid.boundaries.open["id"].tolist()
#         if tmp_args.nudge_temp is None:
#             if namespace.vgrid.is2D():
#                 tmp_args.nudge_temp = {bnd_id: False for bnd_id in bnd_ids}
#             else:
#                 tmp_args.nudge_temp = {bnd_id: True for bnd_id in bnd_ids}
#         elif tmp_args.nudge_temp is True:
#             tmp_args.nudge_temp = {bnd_id: True for bnd_id in bnd_ids}
#         elif tmp_args.nudge_temp is False:
#             tmp_args.nudge_temp = {bnd_id: False for bnd_id in bnd_ids}
#         else:
#             remaining_bounds = list(
#                 set(list(tmp_args.nudge_temp.keys())) - set(bnd_ids)
#             )
#             if np.all(list(tmp_args.nudge_temp.values())) is True:
#                 for bnd_id in remaining_bounds:
#                     tmp_args.nudge_temp.setdefault(bnd_id, False)
#             else:
#                 for bnd_id in remaining_bounds:
#                     tmp_args.nudge_temp.setdefault(bnd_id, True)
#         remaining_bounds = list(set(tmp_args.nudge_temp.keys()) - set(bnd_ids))
#         if len(remaining_bounds) > 0:
#             f = ["No nudging boundaries found with ids:"]
#             for bnd in remaining_bounds:
#                 f.append(bnd)
#             raise ValueError(" ".join(f))
#         return tmp_args.nudge_temp
# 
#     class Itetype4Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             tmp_parser = argparse.ArgumentParser(add_help=False)
#             add_nudge_to_parser(tmp_parser)
#             tmp_args = tmp_parser.parse_known_args()[0]
#             if namespace.hgrid is not None:
#                 nudge_bounds = get_nudge_bnds(namespace)
#                 itetype4 = bctides.itetype.Itetype4(
#                     nudge=nudge_bounds["1"],
#                     rlmax=tmp_args.rlmax_temp,
#                     rnu_day=tmp_args.rnu_day_temp,
#                 )
#                 setattr(namespace, self.dest, itetype4)
# 
#     itetype_group = parser.add_argument_group(
#         "Temperature boundary condition options"
#     ).add_mutually_exclusive_group()
#     # itetype_group.add_argument(
#     #     "--itetype",
#     #     dest="itetype",
#     #     # nargs=1,  # 0 or more
#     #     # const={},
#     #     # action=CustomBoundaryAction,
#     # )
#     itetype_group.add_argument(
#         "--itetype-1",
#         "--temperature-time-history",
#         dest="itetype",
#         # type=itetype.Itetype1
#     )
#     itetype_group.add_argument(
#         "--itetype-2",
#         "--constant-temperature-value",
#         dest="itetype",
#         # type=itetype.Itetype2
#     )
# 
#     itetype_group.add_argument(
#         "--itetype-3",
#         "--temperature-initial-condition",
#         dest="itetype",
#         # type=itetype.Itetype3,
#     )
# 
#     itetype_group.add_argument(
#         "--itetype-4",
#         "--temp-3d",
#         action=Itetype4Action,
#         dest="itetype",
#         nargs=0,
#         const=bctides.itetype.Itetype4,
#     )
# 
# 
# def add_isatype_to_parser(parser):
#     def get_nudge_bnds(namespace):
#         tmp_parser = argparse.ArgumentParser(add_help=False)
#         add_baroclinic_database_to_parser(tmp_parser)
#         add_nudge_to_parser(tmp_parser)
#         tmp_args = tmp_parser.parse_known_args()[0]
#         bnd_ids = namespace.hgrid.boundaries.open["id"].tolist()
#         if tmp_args.nudge_salt is None:
#             if namespace.vgrid.is2D():
#                 tmp_args.nudge_salt = {bnd_id: False for bnd_id in bnd_ids}
#             else:
#                 tmp_args.nudge_salt = {bnd_id: True for bnd_id in bnd_ids}
#         elif tmp_args.nudge_salt is True:
#             tmp_args.nudge_salt = {bnd_id: True for bnd_id in bnd_ids}
#         elif tmp_args.nudge_salt is False:
#             tmp_args.nudge_salt = {bnd_id: False for bnd_id in bnd_ids}
#         else:
#             remaining_bounds = list(
#                 set(list(tmp_args.nudge_salt.keys())) - set(bnd_ids)
#             )
#             if np.all(list(tmp_args.nudge_salt.values())) is True:
#                 for bnd_id in remaining_bounds:
#                     tmp_args.nudge_salt.setdefault(bnd_id, False)
#             else:
#                 for bnd_id in remaining_bounds:
#                     tmp_args.nudge_salt.setdefault(bnd_id, True)
#         remaining_bounds = list(set(tmp_args.nudge_salt.keys()) - set(bnd_ids))
#         if len(remaining_bounds) > 0:
#             f = ["No nudging boundaries found with ids:"]
#             for bnd in remaining_bounds:
#                 f.append(bnd)
#             raise ValueError(" ".join(f))
#         return tmp_args.nudge_salt
# 
#     class Isatype4Action(argparse.Action):
#         def __call__(self, parser, namespace, values, option_string=None):
#             tmp_parser = argparse.ArgumentParser(add_help=False)
#             add_nudge_to_parser(tmp_parser)
#             tmp_args = tmp_parser.parse_known_args()[0]
#             if namespace.hgrid is not None:
#                 nudge_bounds = get_nudge_bnds(namespace)
#                 isatype4 = bctides.isatype.Isatype4(
#                     nudge=nudge_bounds["1"],
#                     rlmax=tmp_args.rlmax_salt,
#                     rnu_day=tmp_args.rnu_day_salt,
#                 )
#                 setattr(namespace, self.dest, isatype4)
# 
#     isatype_group = parser.add_argument_group(
#         "Salinity boundary condition options"
#     ).add_mutually_exclusive_group()
#     # isatype_group.add_argument(
#     #     "--isatype",
#     #     dest="isatype",
#     #     # nargs=1,  # 0 or more
#     #     # const={},
#     #     # action=CustomBoundaryAction,
#     # )
#     isatype_group.add_argument(
#         "--isatype-1",
#         "--salt-th",
#         dest="isatype",
#         # type=isatype.Isatype1
#     )
#     isatype_group.add_argument(
#         "--isatype-2",
#         "--salt-val",
#         dest="isatype",
#         # type=isatype.Isatype2
#     )
# 
#     isatype_group.add_argument(
#         "--isatype-3",
#         # "--salt-ic",
#         dest="isatype",
#         # type=isatype.Isatype3,
#     )
# 
#     isatype_group.add_argument(
#         "--isatype-4",
#         "--salt-3d",
#         action=Isatype4Action,
#         dest="isatype",
#         nargs=0,
#         const=bctides.isatype.Isatype4,
#     )
# 
# 
# class NudgeAction(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
#         nudge = {}
#         if "disable" in option_string:
#             for disable_bnd in values:
#                 nudge.setdefault(disable_bnd, False)
#             if len(nudge) == 0:
#                 nudge = False
#         else:
#             for enable_bnd in values:
#                 nudge.setdefault(enable_bnd, True)
#             if len(nudge) == 0:
#                 nudge = True
#         setattr(namespace, self.dest, nudge)
# 
# 
# def add_temperature_nudge_to_parser(parser):
#     temp_group = parser.add_argument_group("Nudge options for temperature")
#     nudge_parser = temp_group.add_mutually_exclusive_group()
#     nudge_parser.add_argument(
#         "--nudge-temp",
#         "--nudge-temperature",
#         "--nudge-t",
#         dest="nudge_temp",
#         nargs="*",
#         action=NudgeAction,
#         help="Enables temperature nudging. If no arguments are given, the nudging "
#         "is enabled for all the boundaries. If a per-boundary nudging is desired, "
#         "you can pass the boundary id's of the boundaries where nudging should "
#         "be enabled, for example passing --nudge-temp 1 2 4 will enable nudging "
#         "for boundaries with id's 1, 2 and 4, and will disable the nudging "
#         "for the remaining boundaries. Note: This option is mutually exclusive "
#         "to --disable-nudge-temperature.",
#     )
#     nudge_parser.add_argument(
#         "--disable-nudge-temp",
#         "--disable-nudge-temperature",
#         "--disable-nudge-t",
#         dest="nudge_temp",
#         nargs="*",
#         action=NudgeAction,
#         help="Disables temperature nudging. If no arguments are given, the nudging "
#         "is disabled for all the boundaries. If per-boundary nudging disabling is desired, "
#         "you can pass the boundary id's of the boundaries where nudging should "
#         "be disabled, for example passing --disable-nudge-temp 1 2 4 will disable nudging "
#         "for boundaries with id's 1, 2 and 4, and will enable the nudging "
#         "for the remaining boundaries. Note: This option is mutually exclusive "
#         "to --nudge-temperature.",
#     )
#     temp_group.set_defaults(nudge_temp=None)
#     temp_group.add_argument("--rlmax-temp", default=1.5, type=float)
#     temp_group.add_argument("--rnu_day-temp", default=0.25, type=float)
# 
# 
# def add_salinity_nudge_to_parser(parser):
#     salt_group = parser.add_argument_group("Nudge options for salinity")
#     salt_group.add_argument(
#         "--nudge-salt",
#         "--nudge-salinity",
#         "--nudge-s",
#         dest="nudge_salt",
#         nargs="?",
#         action=NudgeAction,
#     )
#     salt_group.add_argument(
#         "--disable-nudge-salt",
#         "--disable-nudge-salinity",
#         "--disable-nudge-s",
#         dest="nudge_salt",
#         nargs="?",
#         action=NudgeAction,
#     )
#     salt_group.set_defaults(nudge_salt=None)
#     salt_group.add_argument("--rlmax-salt", default=1.5, type=float)
#     salt_group.add_argument("--rnu_day-salt", default=0.25, type=float)
# 
# 
# def add_nudge_to_parser(parser):
#     add_temperature_nudge_to_parser(parser)
#     add_salinity_nudge_to_parser(parser)
# 
# 
# def add_ibctype_to_parser(parser):
#     add_iettype_to_parser(parser)
#     add_ifltype_to_parser(parser)
#     add_itetype_to_parser(parser)
#     add_isatype_to_parser(parser)
#     # add_itrtype_to_parser(parser)
#     add_nudge_to_parser(parser)
# 
# 
# def add_windrot_to_parser(parser):
#     parser.add_argument(
#         "--windrot",
#         # action=WindrotAction,
#     )


def add_sflux_options_to_parser(parser):
    parser.add_argument(
        "--file-interval-sflux-1", type=lambda x: timedelta(hours=int(x))
    )
    parser.add_argument(
        "--file-interval-sflux-2", type=lambda x: timedelta(hours=int(x))
    )


class NWS2Action(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) > 2:
            raise ValueError(
                "pyschism forecast init: error: argument --nws-2/--sflux: "
                "expected at most two arguments."
            )
        # if len(values) == 1:
        #     values.append(None)
        sflux_1 = Sflux1Types[values[0].upper()].value(
            product="gfs_0p25_1hr" if values[0].lower() == "gfs" else values[0].lower()
        )
        if len(values) == 2:
            sflux_2 = Sflux2Types[values[1].upper()].value()
        else:
            sflux_2 = None

        tmp_parser = argparse.ArgumentParser(add_help=False)
        if not bool(set(sys.argv).intersection(["-h", "--help"])):
            add_windrot_to_parser(tmp_parser)
            tmp_args = tmp_parser.parse_known_args()[0]
            windrot = (
                gridgr3.Windrot.default(namespace.hgrid)
                if tmp_args.windrot is None
                else tmp_args.windrot
            )
        else:
            windrot = None
        setattr(namespace, self.dest, nws.NWS2(sflux_1, sflux_2, windrot=windrot))


def add_sflux_to_parser(parser):
    parser.add_argument(
        "sflux",
        nargs="+",
        action=NWS2Action,
        metavar="sflux_level",
    )
    add_sflux_options_to_parser(parser)
    #add_windrot_to_parser(parser)


def add_nws_to_parser(parser):

    nws_group = parser.add_argument_group("Atmospheric forcing options")
    nws_parser = nws_group.add_mutually_exclusive_group()
    nws_parser.add_argument(
        "--nws-1",
        dest="nws",
    )
    nws_parser.add_argument(
        "--nws-2",
        "--sflux",
        dest="nws",
        nargs="+",
        action=NWS2Action,
        metavar="sflux_level",
    )
    add_sflux_options_to_parser(parser)
    #add_windrot_to_parser(parser)
    nws_parser.add_argument(
        "--nws-3",
        dest="nws",
    )
    nws_parser.add_argument(
        "--nws-4",
        dest="nws",
    )
    # nws_group.add_argument('--air')
    # nws_group.add_argument('--prc')
    # nws_group.add_argument('--rad')


def add_source_sink_to_parser(parser):
    class SourceSinkAction(argparse.Action):
        class SourceSinkTypes(Enum):
            NWM = source_sink.NWM

            @classmethod
            def _missing_(cls, name):
                f = [
                    f"{name} is not a valid source_sink type. Valid values are: ",
                ]
                for enum_type in cls:
                    f.append(enum_type.name.lower())
                f.append(".")
                raise ValueError("".join(f))

        # def __init__(self, option_strings, dest, nargs=None, **kwargs):
        #     if nargs is not None:
        #         raise ValueError("nargs not allowed")
        #     super().__init__(option_strings, dest, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            ss = source_sink.SourceSink()

            for source_sink_request in values:
                ss.add_forcing(
                    self.SourceSinkTypes[source_sink_request.upper()].value()
                )

            setattr(namespace, self.dest, ss)

    ss_parser = parser.add_argument_group(
        "Source/Sink forcing options"
    ).add_mutually_exclusive_group()
    ss_parser.add_argument(
        "--source-sink",
        dest="source_sink",
        nargs="+",
        action=SourceSinkAction,
        help="Add source/sink terms to model.",
    )


def add_waves_to_parser(parser):
    raise NotImplementedError("add_waves_to_parser")


def add_generic_ic_to_parser(parser, name):

    # class GenericIcAction(argparse.Action):
    #     def __call__(self, parser, namespace, values, option_string=None):
    #         raise NotImplementedError(f'{option_string} is not implemented')
    #         setattr(namespace, self.dest, values)

    ic = parser.add_argument_group(f"{name.capitalize()} initial condition options.")

    # tmp_parser = argparse.ArgumentParser(add_help=False)
    # tmp_parser.add_argument(f'--{name}_ic')
    # add_source_sink_to_parser(tmp_parser)
    # tmp_args = tmp_parser.parse_known_args()[0]
    # if tmp_args.args.elev_ic is None and tmp_parser.source_sink is not None:
    #     default = gridgr3.ElevIc.default(self.hgrid)

    # if self.args.temp_ic is None and self.bctides.itetype is not None:
    #     self.args.temp_ic = gridgr3.TempIc.from_hycom(
    #         self.hgrid,
    #         # self.bctides.itetype.data_source,
    #         self.start_date,
    #     )

    # if self.args.salt_ic is None and self.bctides.isatype is not None:
    #     self.args.salt_ic = gridgr3.SaltIc.from_hycom(
    #         self.hgrid,
    #         # self.bctides.isatype.data_source,
    #         self.start_date,
    #     )

    ic.add_argument(
        f"--{name}-ic",
        # type=gr3type.open,
        # action=GenericIcAction,
        help=f"{name.capitalize()} initial condition file.",
        # default=default,
    )
    ic.add_argument(f"--{name}-ic-crs")


def add_dates_to_parser(parser):
    def add_datetime_format_to_parser(parser):
        parser.add_argument(
            "--datetime-format", "--strptime", default=r"%Y-%m-%dT%H:%M:%S"
        )

    def add_start_date_to_parser(parser):
        class StartDateAction(argparse.Action):
            def __call__(self, parser, namespace, values, option_string=None):
                tmp_parser = argparse.ArgumentParser(add_help=False)
                add_datetime_format_to_parser(tmp_parser)
                tmp_args = tmp_parser.parse_known_args()[0]
                start_date = datetime.strptime(values, tmp_args.datetime_format)
                setattr(namespace, self.dest, start_date)

        parser.add_argument(
            "--start-date",
            action=StartDateAction,
            default=dates.nearest_cycle(),
        )

    def add_end_date_to_parser(parser):
        class EndDateAction(argparse.Action):
            def __call__(self, parser, namespace, values, option_string=None):
                if not isinstance(values, timedelta):
                    tmp_parser = argparse.ArgumentParser(add_help=False)
                    add_datetime_format_to_parser(tmp_parser)
                    tmp_args = tmp_parser.parse_known_args()[0]
                    values = datetime.strptime(values, tmp_args.datetime_format)
                setattr(namespace, self.dest, values)

        end_time_group = parser.add_mutually_exclusive_group(required=True)

        end_time_group.add_argument("--end-date", action=EndDateAction, dest="end_date")

        end_time_group.add_argument(
            "--run-days",
            action=EndDateAction,
            dest="end_date",
            type=lambda x: timedelta(days=float(x)),
        )

    add_start_date_to_parser(parser)
    add_end_date_to_parser(parser)


def add_elev_ic_to_parser(parser):
    class ElevIcAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):

            if "no" in option_string:
                setattr(namespace, self.dest, False)
                return
            # if user passes argument, it will be opened.
            elif values is not None:
                tmp_parser = argparse.ArgumentParser(add_help=False)
                tmp_parser.add_argument("--elev-ic-crs")
                tmp_parser.add_argument("--hgrid-crs")
                tmp_args = tmp_parser.parse_known_args()[0]
                tmp_args.elev_ic_crs = (
                    tmp_args.hgrid_crs
                    if tmp_args.elev_ic_crs is None and tmp_args.hgrid_crs is not None
                    else tmp_args.elev_ic_crs
                )
                elev_ic = gridgr3.ElevIc.open(values, crs=tmp_args.elev_ic_crs)
            else:
                elev_ic = gridgr3.ElevIc.default(namespace.hgrid)
            setattr(namespace, self.dest, elev_ic)

    elev_group = parser.add_argument_group("Elevation initial condition options.")
    elev_ic = elev_group.add_mutually_exclusive_group()
    elev_ic.add_argument(
        "--elev-ic",
        # type=Elev,
        action=ElevIcAction,
        help="Elevation initial condition file. If omitted, one will be created.",
        nargs="?",
        dest="elev_ic",
    )
    elev_ic.add_argument(
        "--no-elev-ic",
        # type=Elev,
        action=ElevIcAction,
        help="Elevation initial condition file. If omitted, one will be created.",
        nargs=0,
        dest="elev_ic",
    )
    elev_group.add_argument(
        "--elev-ic-crs",
        help="Elevation initial condition CRS for grid file. If omitted, it will be assumed that this file "
        "is in the same CRS as the hgrid.",
    )


def add_temp_ic_to_parser(parser):
    add_generic_ic_to_parser(parser, "temp")


def add_salt_ic_to_parser(parser):
    add_generic_ic_to_parser(parser, "salt")


def add_ic_to_parser(parser):
    add_elev_ic_to_parser(parser)
    add_temp_ic_to_parser(parser)
    add_salt_ic_to_parser(parser)


def add_surface_outputs_to_parser(parser):
    def hydro():
        """
        hydro output options
        """
        return {
            1: ("elev", "0: off; 1: on - elev. [m]"),
            2: ("air_pressure", "air pressure [Pa]"),
            3: ("air_temperature", "air temperature [C]"),
            4: ("specific_humidity", "Specific humidity [-]"),
            5: ("solar_radiation", "solar (shortwave) radiation [W/m/m]"),
            6: ("sensible_flux", "sensible flux (positive upward) [W/m/m]"),
            7: ("latent_heat", "latent heat flux (positive upward) [W/m/m]"),
            8: (
                "upward_longwave",
                "upward longwave radiation (positive upward) [W/m/m]",
            ),
            9: (
                "downward_longwave",
                "downward longwave radiation (positive downward) [W/m/m]",
            ),
            10: ("total_heat_flux", "total flux=-flsu-fllu-(radu-radd) [W/m/m]"),
            11: ("evaporation", "evaporation rate [kg/m/m/s]"),
            12: ("precipitation", "precipitation rate [kg/m/m/s]"),
            13: ("bottom_stress", "Bottom stress vector [kg/m/s^2(Pa)]"),
            14: ("wind_speed", "wind velocity vector [m/s]"),
            15: ("wind_stress", "wind stress vector [m^2/s/s]"),
            16: ("dahv", "depth-averaged vel vector [m/s]"),
            17: ("vertical_velocity", "vertical velocity [m/s]"),
            18: ("temp", "water temperature [C]"),
            19: ("salt", "water salinity [PSU]"),
            20: ("water_density", "water density [kg/m^3]"),
            21: ("diffusivity", "eddy diffusivity [m^2/s]"),
            22: ("viscosity", "eddy viscosity [m^2/s]"),
            23: ("TKE", "turbulent kinetic energy"),
            24: ("mixing-lenght", "turbulent mixing length [m]"),
            25: ("hvel", "horizontal vel vector [m/s]"),
            26: ("hvel_side", "horizontal vel vector defined @side [m/s]"),
            27: ("wvel_elem", "vertical vel. @elem [m/s]"),
            28: ("temp_elem", "T @prism centers [C]"),
            29: ("salt_elem", "S @prism centers [PSU]"),
            30: (
                "pressure_gradient",
                "Barotropic pressure gradient force vector (m.s-2) @side " "centers",
            ),
        }

    def wwm():
        """
        WWM output options
        """
        return {
            1: ("WWM_1", "sig. height (m)"),
            2: ("WWM_2", "Mean average period (sec) - TM01"),
            3: (
                "WWM_3",
                "Zero down crossing period for comparison with buoy (s) - TM02",
            ),
            4: ("WWM_4", "Average period of wave runup/overtopping - TM10"),
            5: ("WWM_5", "Mean wave number (1/m)"),
            6: ("WWM_6", "Mean wave length (m)"),
            7: (
                "WWM_9",
                "Mean average energy transport direction (degr) - MWD in NDBC?",
            ),
            8: ("WWM_10", "Mean directional spreading (degr)"),
            9: ("WWM_11", "Discrete peak period (sec) - Tp"),
            10: (
                "WWM_12",
                "Continuous peak period based on higher order moments (sec)",
            ),
            11: ("WWM_13", "Peak phase vel. (m/s)"),
            12: ("WWM_14", "Peak n-factor."),
            13: ("WWM_15", "Peak group vel. (m/s)"),
            14: ("WWM_16", "Peak wave number"),
            15: ("WWM_17", "Peak wave length"),
            16: ("WWM_18", "Peak (dominant) direction (degr)"),
            17: ("WWM_19", "Peak directional spreading"),
            18: ("WWM_20", "Discrete peak direction (radian?) "),
            19: ("WWM_21", "Orbital vel. (m/s) "),
            20: ("WWM_22", "RMS Orbital vel. (m/s) "),
            21: ("WWM_23", "Bottom excursion period (sec?) "),
            22: ("WWM_24", "Bottom wave period (sec) "),
            23: ("WWM_25", "Uresell number based on peak period "),
            24: ("WWM_26", "Friction velocity (m/s?) "),
            25: ("WWM_27", "Charnock coefficient "),
            26: ("WWM_28", "Rougness length "),
            27: ("WWM_energy_dir", "WWM_energy vector"),
            28: (
                "wave-force",
                "Wave force vector (m.s-2) computed by wwm @side centers and "
                "whole levels",
            ),
        }

    def gen():
        """
        gen output options
        """
        return {
            1: ("GEN_1", "1st tracer"),
            2: ("GEN_2", "2nd tracer"),
        }

    def age():
        """
        age output options
        """
        return {
            1: ("AGE_1", 'Indices from "1" to "ntracer_age/2"; [days]'),
            2: ("AGE_2", 'Indices from "1" to "ntracer_age/2"; [days]'),
        }

    def sed():
        """
        sed output options
        """
        return {
            1: ("SED_depth_change", "bottom depth _change_ from init. condition (m)"),
            2: ("SED_D50", " Bed median grain size in the active layer (mm)"),
            3: ("SED_bed_stress", " Bottom shear stress (Pa)"),
            4: ("SED_bed_roughness", " Bottom roughness lenghth (mm)"),
            5: ("SED_TSC", "total suspended concentration (g/L)"),
            6: ("bed_thickness", " total bed thickness @elem (m)"),
            7: ("bed_age", " total bed age over all layers @elem (sec)"),
            8: ("z0st", " Sediment transport roughness length @elem (m) (z0st_elem)"),
            9: ("z0cr", "current-ripples roughness length @elem (m) (z0cr_elem)"),
            10: ("z0sw", "sand-waves roughness length (m) @elem (z0sw_elem)"),
            11: ("z0wr", "wave-ripples roughness length @elem (m) (z0wr_elem)"),
            12: ("SED3D_1", "conc. of 1st class (one output need by each class) [g/L]"),
            13: (
                "SED_bdld_1",
                "Bedload transport rate vector (kg.m-1.s-1) for 1st tracer (one "
                "output need by tracer)",
            ),
            14: (
                "SED_bedfrac_1",
                "Bed fraction 1st tracer (one output need by each class) [-]",
            ),
            15: ("SED3D_2", "conc. of 2nd class"),
            16: ("SED_bdld_2", "Bedload transport of 2nd class"),
            17: ("SED_bedfrac_3", "Bed fraction of 2nd class"),
        }

    def eco():
        """
        EcoSim output options
        """
        return {1: ("ECO_1", "EcoSim outputs")}

    def icm():
        """
        ICM output options
        """
        return {
            1: ("ICM_Chl", "Chlorophyll"),
            2: ("ICM_pH", "PH values (ICM_PH on)"),
            3: ("ICM_PrmPrdt", "ICM primary production @elem [gC/m^3/day]"),
            4: ("ICM_DIN", "ICM totoal inorganic nitrogen (DIN) @elem [gN/m^3]"),
            5: ("ICM_PON", "ICM paticulate organic nitrogen (PON) @elem [gN/m^3]"),
            6: (
                "ICM_SED_BENDOC",
                "ICM bed sediment flux arrays: SED_BENDOC (output "
                "name:ICM_SED_BENDOC) @elem [gC/(m^2 day)]",
            ),
            7: (
                "ICM_SED_BENNH4",
                "ICM bed sediment flux arrays: SED_BENNH4 (output "
                "name:ICM_SED_BENNH4) @elem [gC/(m^2 day)]",
            ),
            8: (
                "ICM_SED_BENNO3",
                "ICM bed sediment flux arrays: SED_BENNO3 (output "
                "name:ICM_SED_BENNO3)@elem [gC/(m^2 day)]",
            ),
            9: (
                "ICM_SED_BENPO4",
                "ICM bed sediment flux arrays: SED_BENPO4 (output "
                "name:ICM_SED_BENPO4) @elem [gC/(m^2 day)]",
            ),
            10: (
                "ICM_SED_BENCOD",
                "ICM bed sediment flux arrays: SED_BENCOD (output "
                "name:ICM_SED_BENCOD) @elem [gC/(m^2 day)]",
            ),
            11: (
                "ICM_SED_BENDO",
                "ICM bed sediment flux arrays: SED_BENDO (output "
                "name:ICM_SED_BENDO) @elem [gC/(m^2 day)]",
            ),
            12: (
                "ICM_SED_BENSA",
                "ICM bed sediment flux arrays: SED_BENSA (output "
                "name:ICM_SED_BENSA) @elem [gC/(m^2 day)]",
            ),
            13: ("ICM_lfsav", "ICM SAV leaf biomass @elem [gC/m^3] (k=1 is surface)"),
            14: ("ICM_stsav", "ICM SAV stem biomass @elem [gC/m^3]"),
            15: ("ICM_rtsav", "ICM SAV root biomass @elem [gC/m^3]"),
            16: ("ICM_tlfsav", "ICM SAV total leaf biomass @elem [gC/m^2]"),
            17: ("ICM_tstsav", "ICM SAV total stem biomass @elem [gC/m^2]"),
            18: ("ICM_trtsav", "ICM SAV total root biomass @elem [gC/m^2]"),
            19: ("ICM_hcansav", "ICM SAV canopy height @elem [m]"),
            20: ("ICM_CNH4", "bottom NH4 conc"),
            21: ("ICM_CNH3", "bottom NO3 conc"),
            22: ("ICM_CPIP", "bottom P conc"),
            23: ("ICM_CPOS", "bottom Si conc"),
            24: ("ICM_CCH4", "bottom CH4 conc"),
            25: ("ICM_CSO4", "bottom SO4 conc"),
            26: ("ICM_CH2S", "bottom H2S conc"),
            27: ("ICM_SEDPON1", "bottom PON g1 conc"),
            28: ("ICM_SEDPON2", "bottom PON g2 conc"),
            29: ("ICM_SEDPON3", "bottom PON g3 conc"),
            30: ("ICM_SEDPOP1", "bottom POP g1 conc"),
            31: ("ICM_SEDPOP2", "bottom POP g2 conc"),
            32: ("ICM_SEDPOP3", "bottom POP g3 conc"),
            33: ("ICM_SEDPOC1", "bottom POC g1 conc"),
            34: ("ICM_SEDPOC2", "bottom POC g2 conc"),
            35: ("ICM_SEDPOC3", "bottom POC g3 conc"),
            36: ("ICM_EROH2S", "erosion flux H2S"),
            37: ("ICM_EROLPOC", "erosion flux LPOC"),
            38: ("ICM_ERORPOC", "erosion flux RPOC"),
            39: ("ICM_DO_consumption", "DO consumption"),
            40: ("ICM_GP1", "PB growth #1"),
            41: ("ICM_GP2", "PB growth #2"),
            42: ("ICM_GP3", "PB growth #3"),
            43: ("ICM_1", "Zoo. #1"),
            44: ("ICM_2", "Zoo. #2"),
            45: ("ICM_3", "phyto #1"),
            46: ("ICM_4", "phyto #2"),
            47: ("ICM_5", "phyto #3"),
            48: ("ICM_6", "RPOC"),
            49: ("ICM_7", "LPOC"),
            50: ("ICM_8", "DOC"),
            51: ("ICM_9", "RPON"),
            52: ("ICM_10", "LPON"),
            53: ("ICM_11", "DON"),
            54: ("ICM_12", "NH4"),
            55: ("ICM_13", "NO3"),
            56: ("ICM_14", "RPOP"),
            57: ("ICM_15", "LPOP"),
            58: ("ICM_16", "DOP"),
            59: ("ICM_17", "PO4t"),
            60: ("ICM_18", "Si- biogenic"),
            61: ("ICM_19", "available Si"),
            62: ("ICM_20", "COD: Chemical oxygen demand"),
            63: ("ICM_21", "DO"),
            64: ("ICM_22", "TIC"),
            65: ("ICM_23", "ALK"),
            66: ("ICM_24", "CA"),
            67: ("ICM_25", "CACO3"),
        }

    def cos():
        """
        CoSINE output options
        """
        return {
            1: ("COS_1", "COS_1"),
            2: ("COS_2", "COS_2"),
            3: ("COS_3", "COS_3"),
            4: ("COS_4", "COS_4"),
            5: ("COS_5", "COS_5"),
            6: ("COS_6", "COS_6"),
            7: ("COS_7", "COS_7"),
            8: ("COS_8", "COS_8"),
            9: ("COS_9", "COS_9"),
            10: ("COS_10", "COS_10"),
            11: ("COS_11", "COS_11"),
            12: ("COS_12", "COS_12"),
            13: ("COS_13", "COS_13"),
        }

    def fib():
        """
        Fecal indicating bacteria output options
        """
        return {1: ("FIB_1", "FIB_1")}

    def sed2d():
        """
        SED2D output options
        """
        return {
            1: ("SED2D_depth_change", "bottom depth _change_ from init. condition (m)"),
            2: ("SED2D_Cd", "drag coefficient used in transport formulae"),
            3: ("SED2D_cflsed", "Courant number (b.qtot.dt / h.dx)"),
            4: ("SED2D_d50", "Top layer d50 (m)"),
            5: ("SED2D_total_transport", "total transport rate vector (kg/m/s)"),
            6: ("SED2D_susp_load", "suspended tranport rate vector (kg/m/s)"),
            7: ("SED2D_bed_load", "bedload transport rate vector (kg/m/s)"),
            8: (
                "SED2D_average_transport",
                "time averaged total transport rate vector (kg/m/s)",
            ),
            9: ("SED2D_bottom_slope", "bottom slope vector (m/m); negative uphill"),
            10: ("z0eq2d", "Total roughness length @elem (m) (z0eq)"),
            11: ("z0cr2d", "current-ripples roughness length @elem (m) (z0cr)"),
            12: ("z0sw2d", "sand-waves roughness length @elem (m) (z0sw)"),
            13: ("z0wr2d", "wave-ripples roughness length @elem (m) (z0wr)"),
        }

    def mar():
        """
        marsh output options
        """
        return {
            1: ("marsh_flag", "marsh_flag"),
        }

    def ice():
        """
        ice output options
        """
        return {
            1: ("ICE_velocity", "ice advective velcoity vector [m/s]"),
            2: ("ICE_strain_rate", "strain rate @ elem [1/sec]"),
            3: ("ICE_net_heat_flux", "net heat flux to ocean (>0 warm up SST) [W/m/m]"),
            4: (
                "ICE_fresh_water_flux",
                "net fresh water flux to ocean (>0 freshens up SSS) [kg/s/m/m]",
            ),
            5: ("ICE_top_T", "ice temperature [C] at air-ice interface"),
            6: ("ICE_tracer_1", "ice volume [m]"),
            7: ("ICE_tracer_2", "ice concentration [-]"),
            8: ("ICE_tracer_3", "snow volume [m]"),
        }

    def ana():
        return {
            1: ("ANA_air_pres_grad_x", "x-component of 𝛁air_pres/ρ0 [m/s/s]"),
            2: ("ANA_air_pres_grad_y", "y-component of 𝛁air_pres/ρ0 [m/s/s]"),
            3: ("ANA_tide_pot_grad_x", "α*g*𝛁Ψ [m/s/s] (gradient of tidal potential)"),
            4: ("ANA_tide_pot_grad_y", "α*g*𝛁Ψ [m/s/s]"),
            5: ("ANA_hor_viscosity_x", "𝛁·(μ𝛁u) [m/s/s] (horizontal viscosity)"),
            6: ("ANA_hor_viscosity_y", "𝛁·(μ𝛁u) [m/s/s]"),
            7: (
                "ANA_bclinic_force_x",
                "-g/rho0* ∫_z^η dr_dx dz  [m/s/s] (b-clinic gradient)",
            ),
            8: ("ANA_bclinic_force_y", "-g/rho0* ∫_z^η dr_dy dz  [m/s/s]"),
            9: (
                "ANA_vert_viscosity_x",
                "d (ν du/dz)/dz [m/s/s] - no vegetation effects (vertical "
                "viscosity)",
            ),
            10: (
                "ANA_vert_viscosity_y",
                "d (ν dv/dz)/dz [m/s/s] - no vegetation effects",
            ),
            11: ("ANA_mom_advection_x", "(u·𝛁) u [m/s/s] (momentum advection)"),
            12: ("ANA_mom_advection_y", "(u·𝛁) u [m/s/s]"),
            13: ("ANA_Richardson", "gradient Richardson number [-]"),
            14: (
                "ANA_transport_min_dt_elem",
                "min time step at each element over all subcycles in horizontal "
                "transport solver [s]  ",
            ),
        }

    surface_outputs_group = parser.add_argument_group("Surface output options")
    surface_outputs_group.add_argument(
        "--nspool",
        help="If passing an integer, it is interpreted as timestep, "
        "if passing a float, it is interpreted as hours.",
    )

    outputs = {
        "hyd": hydro(),
        "wwm": wwm(),
        "gen": gen(),
        "age": age(),
        "sed": sed(),
        "eco": eco(),
        "icm": icm(),
        "cos": cos(),
        "fib": fib(),
        "sed2d": sed2d(),
        "mar": mar(),
        "ice": ice(),
        "ana": ana(),
    }
    for short_name, output in outputs.items():
        for id, (long_name, help_msg) in output.items():
            surface_outputs_group.add_argument(
                f"--{long_name.lower().replace('_', '-')}",
                f"-{short_name}{id}",
                help=help_msg,
                action="store_true",
            )


def add_stations_outputs_to_parser(parser):
    stations_output_group = parser.add_argument_group("Stations output options")
    stations_output_group.add_argument(
        "--stations-file",
        metavar="PATH",
    )
    stations_output_group.add_argument("--stations-file-crs")
    stations_output_group.add_argument(
        "--nspool-sta",
        help="Output interval for stations. If a station file is provided, "
        "and this option is omitted, the stations outputs will be at a six-minute "
        "iterval. If this parameter is provided as a float, the float will be interpreted in minutes, "
        "if an integer is provided, it will be interpreted as iteration step interval.",
    )
    stations_output_group.add_argument(
        "--stations-elev",
        action="store_true",
        help="Include water level output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-prmsl",
        action="store_true",
        help="Include pressure at mean sea level output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-uwind",
        action="store_true",
        help="Include wind u-component output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-vwind",
        action="store_true",
        help="Include wind v-component output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-temp",
        action="store_true",
        help="Include water temperature output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-sal",
        action="store_true",
        help="Include water salinity output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-uvel",
        action="store_true",
        help="Include water u-component output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-vvel",
        action="store_true",
        help="Include water v-component output for stations.",
    )
    stations_output_group.add_argument(
        "--stations-wvel",
        action="store_true",
        help="Include water level output for stations.",
    )


def add_vmin_to_parser(parser):
    parser.add_argument("--vmin", type=float)


def add_vmax_to_parser(parser):
    parser.add_argument("--vmax", type=float)


def add_schism_binary_to_parser(parser):
    parser.add_argument("--schism-binary", default="pschism_TVD-VL")


def add_modules_to_parser(parser):
    parser.add_argument("--modules-init")
    parser.add_argument("--modulepath")
    parser.add_argument("--module", action="append", dest="modules")


def add_slurm_to_parser(parser):
    parser.add_argument("--account")
    parser.add_argument("--partition")
    parser.add_argument(
        "--walltime",
        type=lambda x: TimeDeltaType()(x),  #   help="In hours, float."
    )
    parser.add_argument("--ntasks", required=True, type=int)
    parser.add_argument("--filename")
    parser.add_argument("--rundir")
    parser.add_argument("--run-name")
    parser.add_argument("--mail-type")
    parser.add_argument("--mail-user")
    parser.add_argument("--log-filename")
    parser.add_argument("--path-prepend")
    parser.add_argument("--ld_library_path-prepend")
    parser.add_argument("--nodes")
    parser.add_argument("--launcher", default="srun")
    parser.add_argument("--extra-commands", action="append")


def add_workload_manager_options_to_parser(parser):
    wlm = parser.add_subparsers(dest="workload_manager")
    add_slurm_to_parser(wlm.add_parser("slurm"))

    # --- Parser design #1
    # class WorkloadManagerAction(argparse.Action):

    #     def __call__(self, parser, namespace, values, option_string=None):
    #         setattr(namespace, self.dest, values[2:])
    # options = parser.add_argument_group('Workload manager options')
    # manager = options.add_mutually_exclusive_group()
    # manager.add_argument(
    #     '--torque',
    #     '--pbs',
    #     action=WorkloadManagerAction,
    #     dest='workload_manager',
    #     nargs=0,
    # )
    # manager.add_argument(
    #     '--slurm',
    #     action=WorkloadManagerAction,
    #     dest='workload_manager',
    #     nargs=0,
    # )

    # options.add_argument('--account')
    # options.add_argument('--partition')
    # options.add_argument('--walltime', type=float, help="In hours, float.")
    # options.add_argument('--filename')

    # --- Parser design # 2: Using subparsers
    # options = parser.add_argument_group('Workload manager options')
    # add_torque_to_parser(wlm.add_parser("torque"))
    # add_pbs_to_parser(wlm.add_parser("pbs"))

    # torque = wlm.add_parser("torque")

    # server_config = init.add_subparsers(dest="server_config")
    # slurm = server_config.add_parser(
    #     'slurm', help="Add options for slurm run configuration.")
    # # slurm.add_argument('--account')
    # # slurm.add_argument('--partition')
    # # slurm.add_argument('--walltime', type=float, help="In hours, float.")
    # # slurm.add_argument('--slurm-filename')
    # # slurm.add_argument('--slurm-rundir')
    # # slurm.add_argument('--run-name')
    # # slurm.add_argument('--mail-type')
    # # slurm.add_argument('--mail-user')
    # # slurm.add_argument('--log-filename')
    # slurm.add_argument('--path-prefix')
    # # slurm.add_argument('--slurm-nodes')
    # # slurm.add_argument('--slurm-launcher', default='srun')
    # # slurm.add_argument('--extra-commands', action='append')

    # # add server options
    # parser.add_argument('--hostname')
    # parser.add_argument('--port', type=int)
    # parser.add_argument('--wdir', required=True if '--hostname' in sys.argv else False)
    # parser.add_argument('--keep-wdir', action='store_true')
    # parser.add_argument('--binaries-path', '--binaries-prefix', dest='binaries_prefix')
    # parser.add_argument('--source-script')
    # parser.add_argument('--additional-mpi-options')

    # # make nproc required when using ssh
    # # args = parser.parse_known_args()[0]
    # if '--hostname' in sys.argv:
    #     parser.add_argument('--nproc', '--ncpu', type=int, required=True)
    # else:
    #     parser.add_argument('--nproc', '--ncpu', type=int, default=-1)

    # # add resource manager option
    # manager = parser.add_mutually_exclusive_group()
    # manager.add_argument('--use-torque', action='store_true')
    # manager.add_argument('--use-pbs', action='store_true')
    # manager.add_argument('--use-slurm', action='store_true')

    # # flag some options as required when a resource manager is enabled
    # _required = '--use-torque' in sys.argv
    # _required = _required | ('--use-pbs' in sys.argv)
    # _required = _required | ('--use-slurm' in sys.argv)

    # # resource manager specific options
    # # parser.add_argument('--account', required=_required)
    # parser.add_argument('--slurm-ntasks', required=_required, type=int)
    # # parser.add_argument('--walltime', required=_required, type=float)
    # # parser.add_argument('--partition')
    # # parser.add_argument('--slurm-filename')
    # # parser.add_argument('--slurm-rundir')
    # # parser.add_argument('--run-name')
    # # parser.add_argument('--mail-type')
    # # parser.add_argument('--mail-user')
    # # parser.add_argument('--log-filename')
    # # parser.add_argument('--slurm-nodes')
    # # parser.add_argument('--slurm-launcher', default='srun')
    # # parser.add_argument('--extra-commands', action='append')
    # # parser.add_argument('--module', default=list(), action='append', dest='modules')



from pyschism.forcing.bctides.tides import Tides
# from pyschism.forcing.tides.bctides import Bctides
from pyschism.forcing.nws.nws2.gfs import GlobalForecastSystem
from pyschism.forcing.source_sink.nwm import NationalWaterModel
from pyschism.forcing.nws.base import NWS

GFS = GlobalForecastSystem
NWM = NationalWaterModel

__all__ = [
    "Tides",
    "GlobalForecastSystem", 'GFS',
    'NationalWaterModel', 'NWM',
    'NWS',
    # 'Bctides',
]



from datetime import datetime
from time import time
import pathlib

from pyschism.forcing.nws.nws2.era5 import ERA5

from pyschism.mesh.hgrid import Hgrid

if __name__ == '__main__':

    startdate=datetime(2018, 1, 1)
    rnday=366

    t0=time()
    hgrid=Hgrid.open('./hgrid.gr3',crs='EPSG:4326')
    bbox = hgrid.get_bbox('EPSG:4326', output_type='bbox')

    er=ERA5()
    outdir = pathlib.Path('./')
    interval = 1
    er.write(
        outdir=outdir, 
        start_date=startdate, 
        rnday=rnday, 
        air=True,   #sflux_air_1
        rad=True,   #sflux_rad_1
        prc=True,   #sflux_prc_1
        bbox=bbox, 
        output_interval=interval,  #raw data is hourly
        overwrite=True,
        tmpdir=outdir) #default tmpdir is system's tmp (/tmp), it is possbile /tmp has no enough space for large dataset.

    #write sflux_inputs.txt
    with open("./sflux_inputs.txt", "w") as f:
        f.write("&sflux_inputs\n/\n")

    print(f'It took {(time()-t0)/60} minutes to generate {rnday} days')



from abc import ABC
from collections import defaultdict
from functools import lru_cache
import hashlib
import logging
from itertools import permutations
import os
import pathlib
import tempfile
from typing import Union, Sequence, Hashable, List, Dict

import geopandas as gpd
from matplotlib.collections import PolyCollection
from matplotlib.path import Path
from matplotlib.tri import Triangulation
from matplotlib.transforms import Bbox
import numpy as np
from pyproj import Transformer, CRS
import requests
from shapely import ops
from shapely.geometry import (
    box,
    LinearRing,
    LineString,
    MultiPolygon,
    Polygon,
    Point,
)

from pyschism.mesh.parsers import grd, sms2dm
from pyschism.figures import figure

logger = logging.getLogger(__name__)


class Nodes:
    def __init__(self, nodes: Dict[Hashable, List[List]], crs=None):
        """Setter for the nodes attribute.

        Argument nodes must be of the form:
            {id: [(x0, y0), z0]}
            or
            {id: [(x0, y0), [z0, ..., zn]}

        Gr3 format is assumed to be exclusively a 2D format that can hold
        triangles or quads.

        """

        for coords, _ in nodes.values():
            if len(coords) != 2:
                raise ValueError(
                    "Coordinate vertices for a gr3 type must be 2D, but got "
                    f"coordinates {coords}."
                )

        self._id = list(nodes.keys())
        self._coords = np.array([coords for coords, _ in nodes.values()])
        self._crs = CRS.from_user_input(crs) if crs is not None else crs
        self._values = np.array([value for _, value in nodes.values()])

    def transform_to(self, dst_crs):
        dst_crs = CRS.from_user_input(dst_crs)
        if not self.crs.equals(dst_crs):
            self._coords = self.get_xy(dst_crs)
            self._crs = dst_crs

        if hasattr(self, "_gdf"):
            del self._gdf

    def transform_to_cpp(self, lonc, latc):
        longitude = list(self.coord[:, 0]/180*np.pi)
        latitude = list(self.coord[:, 1]/180*np.pi)
        radius = 6378206.4
        loncc = lonc / 180 * np.pi
        latcc = latc / 180 * np.pi
        x = [
            radius * (longitude[i] - loncc) * np.cos(latcc)
            for i in np.arange(len(longitude))
        ]
        y = [radius * latitude[i] for i in np.arange(len(latitude))]
        self._coords = np.vstack([x, y]).T
        self._crs = None
        return x, y

    def get_xy(self, crs: Union[CRS, str] = None):
        if crs is not None:
            crs = CRS.from_user_input(crs)
            if not crs.equals(self.crs):
                transformer = Transformer.from_crs(self.crs, crs, always_xy=True)
                x, y = transformer.transform(self.coord[:, 0], self.coord[:, 1])
                return np.vstack([x, y]).T
        return self.coord

    @property
    def gdf(self):
        if not hasattr(self, "_gdf"):
            data = []
            for id, coord, values in zip(self._id, self._coords, self.values):
                data.append({"geometry": Point(coord), "id": id, "values": values})
            self._gdf = gpd.GeoDataFrame(data, crs=self.crs)
        return self._gdf

    @property
    def id(self):
        return self._id

    @property
    def index(self):
        if not hasattr(self, "_index"):
            self._index = np.arange(len(self._id))
        return self._index

    @property
    def crs(self):
        return self._crs

    @property
    def values(self):
        return self._values

    @property
    def coords(self):
        return self._coords

    @property
    def coord(self):
        return self.coords

    def get_index_by_id(self, id: Hashable):
        if not hasattr(self, "node_id_to_index"):
            self.node_id_to_index = {self.id[i]: i for i in range(len(self.id))}
        return self.node_id_to_index[id]

    def get_id_by_index(self, index: int):
        if not hasattr(self, "node_index_to_id"):
            self.node_index_to_id = {i: self.id[i] for i in range(len(self.id))}
        return self.node_index_to_id[index]

    def to_dict(self):
        nodes = {
            nid: (coo, val)
            for nid, coo, val in zip(self._id, self._coords, self.values)
        }
        return nodes


class Elements:
    def __init__(self, nodes: Nodes, elements: Dict[Hashable, Sequence]):
        if not isinstance(elements, dict):
            raise TypeError("Argument elements must be a dict.")

        vertex_id_set = set(nodes.id)
        for id, geom in elements.items():
            if not isinstance(geom, Sequence):
                raise TypeError(
                    f"Element with id {id} of the elements "
                    f"argument must be of type {Sequence}, not "
                    f"type {type(geom)}."
                )
            if not set(geom).issubset(vertex_id_set):
                ValueError(
                    f"Element with id {id} is not a subset of the " "coordinate id's."
                )
        self.nodes = nodes
        self.elements = elements

    def __len__(self):
        return len(self.elements)

    def to_dict(self):
        return self.elements

    @property
    def id(self):
        if not hasattr(self, "_id"):
            self._id = list(self.elements.keys())
        return self._id

    @property
    def index(self):
        if not hasattr(self, "_index"):
            self._index = np.arange(len(self.elements))
        return self._index

    def get_index_by_id(self, id: Hashable):
        if not hasattr(self, "element_id_to_index"):
            self.element_id_to_index = {self.id[i]: i for i in range(len(self.id))}
        return self.element_id_to_index[id]

    def get_id_by_index(self, index: int):
        if not hasattr(self, "element_index_to_id"):
            self.element_index_to_id = {i: self.id[i] for i in range(len(self.id))}
        return self.element_index_to_id[index]

    def get_indexes_around_index(self, index):
        if not hasattr(self, "indexes_around_index"):

            def append_geom(geom):
                for simplex in geom:
                    for i, j in permutations(simplex, 2):
                        indexes_around_index[i].add(j)

            indexes_around_index = defaultdict(set)
            append_geom(self.triangles)
            append_geom(self.quads)
            self.indexes_around_index = indexes_around_index
        return list(self.indexes_around_index[index])

    def get_ball(self, order: int, id=None, index=None):

        if not isinstance(order, int):
            raise TypeError("Argument 'order' must be of type int.")

        if not order >= 0:
            raise TypeError("Argument 'order' must be of greater " "than zero.")

        if id is None and index is None:
            raise ValueError("Must specify one keyword argument of index or id.")

        if id is not None and index is not None:
            raise ValueError(
                "Must specify only one keyword argument of " "index or id."
            )

        if id is not None:
            index = self.get_index_by_id(id)

        eidxs = set([index])
        for i in range(order):
            elements = self.array[list(sorted(eidxs)), :]
            new_neighbors = list(
                map(self.get_indexes_around_index, list(set(elements.data.flatten())))
            )
            new_neighbors = set([item for sublist in new_neighbors for item in sublist])
            eidxs = eidxs.union(
                set(
                    np.where(
                        np.logical_and(
                            np.any(
                                np.isin(self.array, list(set(new_neighbors))), axis=1
                            ),
                            np.any(np.isin(self.array, elements), axis=1),
                        )
                    )[0]
                )
            )
        return self.gdf.loc[eidxs].geometry.unary_union.exterior

    def get_node_ball(self):
        '''
        compute nodal ball information
        '''
        elnode = self.array
        NP = len(self.nodes.values)
        nne = np.zeros(NP).astype('int')
        ine = [[] for i in np.arange(NP)]
        mask = np.any(elnode.mask)
        for i, element in enumerate(elnode):
            ele = element[~element.mask]
            if mask:
                i34 = len(ele)
            else:
                i34 = len(ele[0])
            inds = elnode[i, :i34]
            nne[inds] = nne[inds]+1
            [ine[indi].append(i) for indi in inds]
        ine = np.array([np.array(ine[i]) for i in np.arange(NP)], dtype='O') 
        return nne, ine

    def compute_centroid(self):
        elnode = self.array
        NE = self.elements.__len__()
        depth = self.nodes.values

        x_centr, y_centr, dp_centr = np.zeros([3, NE])

        if np.any(elnode.mask):
            mask = elnode.mask[:, -1]
            #centroid for tris 
            x_centr[mask] = self.nodes.coords[elnode[mask, :3], 0].mean(axis=1)
            y_centr[mask] = self.nodes.coords[elnode[mask, :3], 1].mean(axis=1)
            dp_centr[mask] = depth[elnode[mask, :3]].mean(axis=1)

            #centroid for quads
            x_centr[~mask] = self.nodes.coords[elnode[~mask, :], 0].mean(axis=1)
            y_centr[~mask] = self.nodes.coords[elnode[~mask, :], 1].mean(axis=1)
            dp_centr[~mask] = depth[elnode[~mask, :]].mean(axis=1)
        else:
            x_centr = self.nodes.coords[elnode, 0].mean(axis=1)
            y_centr = self.nodes.coords[elnode, 1].mean(axis=1)
            dp_centr = depth[elnode].mean(axis=1)

        return x_centr, y_centr, dp_centr

    def get_areas(self, crs: Union[CRS, str] = None):
        xy = self.nodes.get_xy(crs)
        x = xy[:, 0]
        y = xy[:, 1]

        elnode = self.array
        x1 = x[elnode[:, 0]]; y1 = y[elnode[:, 0]]
        x2 = x[elnode[:, 1]]; y2 = y[elnode[:, 1]]
        x3 = x[elnode[:, 2]]; y3 = y[elnode[:, 2]]
        if np.any(elnode.mask):
            x4 = x[elnode[:, 3]]; y4 = y[elnode[:, 3]]
            mask = elnode.mask[:, -1]
            x4[mask] = x1[mask]; y4[mask] = y1[mask]
        else:
            x4 = x1; y4 = y1
        area=((x2-x1)*(y3-y1)-(x3-x1)*(y2-y1)+(x3-x1)*(y4-y1)-(x4-x1)*(y3-y1))/2
      
        return area
        
        ##self.gdf method is very slow
        #if self.nodes.crs.is_geographic:
        #    elements = []
        #    for row in self.gdf.itertuples():
        #        aeqd = CRS.from_user_input(
        #            f"+proj=aeqd +R=6371000 +units=m "
        #            f"+lat_0={row.geometry.centroid.y} +lon_0={row.geometry.centroid.x}"
        #        )
        #        current_to_aeqd = Transformer.from_crs(
        #            self.nodes.crs, aeqd, always_xy=True
        #        ).transform
        #        elements.append(ops.transform(current_to_aeqd, row.geometry))
        #    return [element.area for element in elements]
        #else:
        #    return [row.geometry.area for row in self.gdf.itertuples()]

    def get_triangulation_mask(self, element_mask):

        triangulation_mask = []
        for i, element in enumerate(self.elements.values()):
            if len(element) == 3:
                if element_mask[i]:
                    triangulation_mask.append(True)
                else:
                    triangulation_mask.append(False)
        for i, element in enumerate(self.elements.values()):
            if len(element) == 4:
                if element_mask[i]:
                    triangulation_mask.append(True)
                    triangulation_mask.append(True)
                else:
                    triangulation_mask.append(False)
                    triangulation_mask.append(False)

        return np.array(triangulation_mask)


    @property
    def array(self):
        if not hasattr(self, "_array"):
            rank = int(max(map(len, self.elements.values())))
            array = np.full((len(self.elements), rank), -1)
            for i, element in enumerate(self.elements.values()):
                row = np.array(list(map(self.nodes.get_index_by_id, element)))
                array[i, : len(row)] = row
            array = np.ma.masked_equal(array, -1)
            self._array = array
        return self._array

    @property
    def i34(self):
        if not hasattr(self, "_i34"):
            if np.any(self.array.mask):
                self._i34 = np.sum(~self.array.mask, axis=1)
            else:
                self._i34 = np.full(self.array.shape[0], 3)
        return self._i34

    @property
    def triangles(self):
        if not hasattr(self, "_triangles"):
            self._triangles = np.array(
                [
                    list(map(self.nodes.get_index_by_id, element))
                    for element in self.elements.values()
                    if len(element) == 3
                ]
            )
        return self._triangles

    @property
    def tri_idxs(self):
        if not hasattr(self, "_tri_idxs"):
            tri_idxs = np.full(len(self.elements), -99, dtype=int)
            for i, element in enumerate(self.elements.values()):
                if len(element) == 3:
                    tri_idxs[i] = i
            new_arr = np.delete(tri_idxs, np.where(tri_idxs == -99))
            self._tri_idxs = new_arr
        return self._tri_idxs

    @property
    def quadrilaterals(self):
        return self.quads

    @property
    def quads(self):
        if not hasattr(self, "_quads"):
            self._quads = np.array(
                [
                    list(map(self.nodes.get_index_by_id, element))
                    for element in self.elements.values()
                    if len(element) == 4
                ]
            )
        return self._quads

    @property
    def qua_idxs(self):
        if not hasattr(self, "_qua_idxs"):
            qua_idxs = np.full(len(self.elements), -99, dtype=int)
            for i, element in enumerate(self.elements.values()):
                if len(element) == 4:
                    qua_idxs[i] = i
            new_arr = np.delete(qua_idxs, np.where(qua_idxs == -99))
            self._qua_idxs = new_arr
        return self._qua_idxs

    @property
    def sides(self):
        if not hasattr(self, "_sides"):

            sides = []
            for element in self.elements.values():
                if len(element) == 3:
                    results = list(map(self.nodes.get_index_by_id, element))
                    # print(len(results))
                    # print(f'nodes are {results[0]}, {results[1]}, {results[2]}')
                    sides.append([results[1], results[2]])
                    sides.append([results[2], results[0]])
                    sides.append([results[0], results[1]])

                elif len(element) == 4:
                    results = list(map(self.nodes.get_index_by_id, element))
                    # print(f'nodes are {p1}, {p2}, {p3}, {p4}')
                    sides.append([results[1], results[2]])
                    sides.append([results[2], results[3]])
                    sides.append([results[3], results[0]])
                    sides.append([results[0], results[1]])

            # from pyPoseidon
            def remove_reversed_duplicates(iterable):
                # Create a set for already seen elements
                seen = set()
                for item in iterable:
                    # Lists are mutable so we need tuples for the set-operations.
                    tup = tuple(item)
                    if tup not in seen:
                        # If the tuple is not in the set append it in REVERSED order.
                        seen.add(tup[::-1])
                        # If you also want to remove normal duplicates uncomment the next line
                        # seen.add(tup)
                        yield item

            self._sides = np.array(list(remove_reversed_duplicates(sides)))

        return self._sides

    @property
    def triangulation(self):
        if not hasattr(self, "_triangulation"):
            triangles = self.triangles.tolist()
            for quad in self.quads:
                triangles.append([quad[0], quad[1], quad[3]])
                triangles.append([quad[1], quad[2], quad[3]])
            self._triangulation = Triangulation(
                self.nodes.coord[:, 0], self.nodes.coord[:, 1], triangles
            )
        return self._triangulation

    @property
    def gdf(self):
        if not hasattr(self, "_gdf"):
            logger.info("Generating elements geodataframe.")
            from time import time

            start = time()
            data = []
            for id, element in self.elements.items():
                data.append(
                    {
                        "geometry": Polygon(
                            self.nodes.coord[list(map(self.get_index_by_id, element))]
                        ),
                        "id": id,
                    }
                )
            self._gdf = gpd.GeoDataFrame(data, crs=self.nodes.crs)
            logger.info(
                "Generating elements geodataframe took " f"{time()-start} seconds."
            )
        return self._gdf


class Edges:
    def __init__(self, grd: "Gr3"):
        self.gr3 = grd

    @lru_cache(maxsize=1)
    def __call__(self) -> gpd.GeoDataFrame:
        data = []
        for ring in self.gr3.hull.rings().itertuples():
            coords = ring.geometry.coords
            for i in range(1, len(coords)):
                data.append(
                    {
                        "geometry": LineString([coords[i - 1], coords[i]]),
                        "bnd_id": ring.bnd_id,
                        "type": ring.type,
                    }
                )
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    def exterior(self):
        return self().loc[self()["type"] == "exterior"]

    def interior(self):
        return self().loc[self()["type"] == "interior"]


class Rings:
    """Note: This only considers split-quads."""

    def __init__(self, grd: "Gr3"):
        self.gr3 = grd

    @lru_cache(maxsize=1)
    def __call__(self) -> gpd.GeoDataFrame:
        data = []
        for bnd_id, rings in self.sorted().items():
            coords = self.gr3.nodes.coord[rings["exterior"][:, 0], :]
            geometry = LinearRing(coords)
            data.append({"geometry": geometry, "bnd_id": bnd_id, "type": "exterior"})
            for interior in rings["interiors"]:
                coords = self.gr3.nodes.coord[interior[:, 0], :]
                geometry = LinearRing(coords)
                data.append(
                    {"geometry": geometry, "bnd_id": bnd_id, "type": "interior"}
                )
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    def exterior(self):
        return self().loc[self()["type"] == "exterior"]

    def interior(self):
        return self().loc[self()["type"] == "interior"]

    @lru_cache(maxsize=1)
    def sorted(self):
        tri = self.gr3.elements.triangulation
        idxs = np.vstack(list(np.where(tri.neighbors == -1))).T
        boundary_edges = []
        for i, j in idxs:
            boundary_edges.append((tri.triangles[i, j], tri.triangles[i, (j + 1) % 3]))
        return sort_rings(edges_to_rings(boundary_edges), self.gr3.nodes.coord)


class Hull:
    def __init__(self, grd: "Gr3"):
        self.gr3 = grd
        self.edges = Edges(grd)
        self.rings = Rings(grd)

    @lru_cache(maxsize=1)
    def __call__(self) -> gpd.GeoDataFrame:
        data = []
        for bnd_id in np.unique(self.rings()["bnd_id"].tolist()):
            exterior = self.rings().loc[
                (self.rings()["bnd_id"] == bnd_id)
                & (self.rings()["type"] == "exterior")
            ]
            interiors = self.rings().loc[
                (self.rings()["bnd_id"] == bnd_id)
                & (self.rings()["type"] == "interior")
            ]
            data.append(
                {
                    "geometry": Polygon(
                        exterior.iloc[0].geometry.coords,
                        [row.geometry.coords for _, row in interiors.iterrows()],
                    ),
                    "bnd_id": bnd_id,
                }
            )
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    @lru_cache(maxsize=1)
    def exterior(self):
        data = []
        for exterior in (
            self.rings().loc[self.rings()["type"] == "exterior"].itertuples()
        ):
            data.append({"geometry": Polygon(exterior.geometry.coords)})
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    @lru_cache(maxsize=1)
    def interior(self):
        data = []
        for interior in (
            self.rings().loc[self.rings()["type"] == "interior"].itertuples()
        ):
            data.append({"geometry": Polygon(interior.geometry.coords)})
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    @lru_cache(maxsize=1)
    def implode(self) -> gpd.GeoDataFrame:
        return gpd.GeoDataFrame(
            {
                "geometry": MultiPolygon(
                    [polygon.geometry for polygon in self().itertuples()]
                )
            },
            crs=self.gr3.crs,
        )

    @lru_cache(maxsize=1)
    def multipolygon(self) -> MultiPolygon:
        polygon_collection = []
        for rings in self.rings.sorted().values():
            exterior = self.gr3.nodes.coord[rings["exterior"][:, 0], :]
            interiors = []
            for interior in rings["interiors"]:
                interiors.append(self.gr3.nodes.coord[interior[:, 0], :])
            polygon_collection.append(Polygon(exterior, interiors))
        mp = MultiPolygon(polygon_collection)
        if isinstance(mp, Polygon):
            mp = MultiPolygon([mp])
        return mp


class Gr3(ABC):
    def __init__(self, nodes, elements=None, description=None, crs=None):

        self.nodes = Nodes(nodes, crs)
        self.elements = Elements(self.nodes, elements)
        self.description = "" if description is None else str(description)
        self.hull = Hull(self)

    def __str__(self):
        return grd.to_string(**self.to_dict())

    def to_dict(self):
        return {
            "description": self.description,
            "nodes": self.nodes.to_dict(),
            "elements": self.elements.elements,
            "crs": self.crs,
        }

    def write(self, path, overwrite=False, format="gr3"):
        if format in ["gr3", "grd"]:
            grd.write(self.to_dict(), path, overwrite)
        elif format in ["sms", "2dm", "sms2dm"]:
            sms2dm.write(
                {
                    "ND": {
                        i
                        + 1: (
                            coord,
                            -self.values[i] if not np.isnan(self.values[i]) else -99999,
                        )
                        for i, coord in enumerate(self.coords)
                    },
                    "E3T": {i + 1: index + 1 for i, index in enumerate(self.triangles)},
                    "E4Q": {i + 1: index + 1 for i, index in enumerate(self.quads)},
                },
                path,
                overwrite,
            )
        else:
            raise ValueError(f"Unknown format {format} for hgrid output.")

    def get_xy(self, crs: Union[CRS, str] = None):
        return self.nodes.get_xy(crs)

    def get_xyz(self, vgrid, crs: Union[CRS, str] = None):
        return vgrid.get_xyz(self)

    def get_bbox(
        self, crs: Union[str, CRS] = None, output_type: str = None
    ) -> Union[Polygon, Bbox]:
        output_type = "bbox" if output_type is None else output_type
        xmin, xmax = np.min(self.coord[:, 0]), np.max(self.coord[:, 0])
        ymin, ymax = np.min(self.coord[:, 1]), np.max(self.coord[:, 1])
        crs = self.crs if crs is None else crs
        if crs is not None:
            if not self.crs.equals(crs):
                transformer = Transformer.from_crs(self.crs, crs, always_xy=True)
                (xmin, xmax), (ymin, ymax) = transformer.transform(
                    (xmin, xmax), (ymin, ymax)
                )
        if output_type == "polygon":
            return box(xmin, ymin, xmax, ymax)
        elif output_type == "bbox":
            return Bbox([[xmin, ymin], [xmax, ymax]])
        else:
            raise TypeError(
                "Argument output_type must a string literal 'polygon' or " "'bbox'"
            )

    def invert_sign(self):
        self.nodes.values[:] = -self.nodes.values

    def transform_to(self, dst_crs):
        """Transforms coordinate system of mesh in-place."""
        self.nodes.transform_to(dst_crs)

    def vertices_around_vertex(self, index):
        return self.nodes.vertices_around_vertex(index)

    def copy(self):
        return self.__class__(**self.to_dict())

    @classmethod
    def open(cls, file: Union[str, os.PathLike], crs: Union[str, CRS] = None):
        if str(file).endswith(".ll") and crs is None:
            crs = "epsg:4326"
        try:
            response = requests.get(file)
            response.raise_for_status()
            tmpfile = tempfile.NamedTemporaryFile()
            with open(tmpfile.name, "w") as fh:
                fh.write(response.text)
            return cls(
                **grd.read(pathlib.Path(tmpfile.name), crs=crs), boundaries=False
            )
        except Exception:
            pass
        return cls(**grd.read(pathlib.Path(file), boundaries=False, crs=crs))

    @figure
    def tricontourf(self, axes=None, show=True, figsize=None, **kwargs):
        if len(self.triangles) > 0:
            axes.tricontourf(self.x, self.y, self.triangles, self.values, **kwargs)
        return axes

    @figure
    def tripcolor(self, axes=None, show=True, figsize=None, **kwargs):
        if len(self.triangles) > 0:
            axes.tripcolor(self.x, self.y, self.triangles, self.values, **kwargs)
        return axes

    @figure
    def triplot(
        self,
        axes=None,
        show=False,
        figsize=None,
        linewidth=0.07,
        color="black",
        **kwargs,
    ):
        if len(self.triangles) > 0:
            kwargs.update({"linewidth": linewidth})
            kwargs.update({"color": color})
            axes.triplot(self.x, self.y, self.triangles, **kwargs)
        return axes

    @figure
    def quadplot(
        self,
        axes=None,
        show=False,
        figsize=None,
        facecolor="none",
        edgecolor="k",
        linewidth=0.07,
        **kwargs,
    ):
        if len(self.quads) > 0:
            pc = PolyCollection(
                self.coords[self.quads],
                facecolor=facecolor,
                edgecolor=edgecolor,
                linewidth=0.07,
            )
            axes.add_collection(pc)
        return axes

    @figure
    def quadface(self, axes=None, show=False, figsize=None, **kwargs):
        if len(self.quads) > 0:
            pc = PolyCollection(self.coords[self.quads], **kwargs)
            quad_value = np.mean(self.values[self.quads], axis=1)
            pc.set_array(quad_value)
            axes.add_collection(pc)
        return axes

    @figure
    def wireframe(self, axes=None, show=False, **kwargs):
        axes = self.triplot(axes=axes, **kwargs)
        axes = self.quadplot(axes=axes, **kwargs)
        return axes

    @property
    def coords(self):
        return self.nodes.coord

    @property
    def coord(self):
        return self.nodes.coord

    @property
    def vertices(self):
        return self.nodes.coord

    @property
    def vertex_id(self):
        return self.nodes.id

    @property
    def element_id(self):
        return self.elements.id

    @property
    def values(self):
        return self.nodes.values

    @property
    def crs(self):
        return self.nodes.crs

    @property
    def x(self):
        return self.nodes.coord[:, 0]

    @property
    def y(self):
        return self.nodes.coord[:, 1]

    @property
    def triangles(self):
        return self.elements.triangles

    @property
    def quads(self):
        return self.elements.quads

    @property
    def triangulation(self):
        return self.elements.triangulation

    @property
    def bbox(self):
        return self.get_bbox()

    @property
    def md5(self):
        return hashlib.md5(str(self).encode()).hexdigest()


def edges_to_rings(edges):
    if len(edges) == 0:
        return edges
    # start ordering the edges into linestrings
    edge_collection = list()
    ordered_edges = [edges.pop(-1)]
    e0, e1 = [list(t) for t in zip(*edges)]
    while len(edges) > 0:
        if ordered_edges[-1][1] in e0:
            idx = e0.index(ordered_edges[-1][1])
            ordered_edges.append(edges.pop(idx))
        elif ordered_edges[0][0] in e1:
            idx = e1.index(ordered_edges[0][0])
            ordered_edges.insert(0, edges.pop(idx))
        elif ordered_edges[-1][1] in e1:
            idx = e1.index(ordered_edges[-1][1])
            ordered_edges.append(list(reversed(edges.pop(idx))))
        elif ordered_edges[0][0] in e0:
            idx = e0.index(ordered_edges[0][0])
            ordered_edges.insert(0, list(reversed(edges.pop(idx))))
        else:
            edge_collection.append(tuple(ordered_edges))
            idx = -1
            ordered_edges = [edges.pop(idx)]
        e0.pop(idx)
        e1.pop(idx)
    # finalize
    if len(edge_collection) == 0 and len(edges) == 0:
        edge_collection.append(tuple(ordered_edges))
    else:
        edge_collection.append(tuple(ordered_edges))
    return edge_collection


def sort_rings(index_rings, vertices):
    """Sorts a list of index-rings.

    Takes a list of unsorted index rings and sorts them into an "exterior" and
    "interior" components. Any doubly-nested rings are considered exterior
    rings.

    TODO: Refactor and optimize. Calls that use :class:matplotlib.path.Path can
    probably be optimized using shapely.
    """

    # sort index_rings into corresponding "polygons"
    areas = list()
    for index_ring in index_rings:
        e0, e1 = [list(t) for t in zip(*index_ring)]
        areas.append(float(Polygon(vertices[e0, :]).area))

    # maximum area must be main mesh
    idx = areas.index(np.max(areas))
    exterior = index_rings.pop(idx)
    areas.pop(idx)
    _id = 0
    _index_rings = dict()
    _index_rings[_id] = {"exterior": np.asarray(exterior), "interiors": []}
    e0, e1 = [list(t) for t in zip(*exterior)]
    path = Path(vertices[e0 + [e0[0]], :], closed=True)
    while len(index_rings) > 0:
        # find all internal rings
        potential_interiors = list()
        for i, index_ring in enumerate(index_rings):
            e0, e1 = [list(t) for t in zip(*index_ring)]
            if path.contains_point(vertices[e0[0], :]):
                potential_interiors.append(i)
        # filter out nested rings
        real_interiors = list()
        for i, p_interior in reversed(list(enumerate(potential_interiors))):
            _p_interior = index_rings[p_interior]
            check = [
                index_rings[k]
                for j, k in reversed(list(enumerate(potential_interiors)))
                if i != j
            ]
            has_parent = False
            for _path in check:
                e0, e1 = [list(t) for t in zip(*_path)]
                _path = Path(vertices[e0 + [e0[0]], :], closed=True)
                if _path.contains_point(vertices[_p_interior[0][0], :]):
                    has_parent = True
            if not has_parent:
                real_interiors.append(p_interior)
        # pop real rings from collection
        for i in reversed(sorted(real_interiors)):
            _index_rings[_id]["interiors"].append(np.asarray(index_rings.pop(i)))
            areas.pop(i)
        # if no internal rings found, initialize next polygon
        if len(index_rings) > 0:
            idx = areas.index(np.max(areas))
            exterior = index_rings.pop(idx)
            areas.pop(idx)
            _id += 1
            _index_rings[_id] = {"exterior": np.asarray(exterior), "interiors": []}
            e0, e1 = [list(t) for t in zip(*exterior)]
            path = Path(vertices[e0 + [e0[0]], :], closed=True)
    return _index_rings


def signed_polygon_area(vertices):
    # https://code.activestate.com/recipes/578047-area-of-polygon-using-shoelace-formula/
    n = len(vertices)  # of vertices
    area = 0.0
    for i in range(n):
        j = (i + 1) % n
        area += vertices[i][0] * vertices[j][1]
        area -= vertices[j][0] * vertices[i][1]
        return area / 2.0



import os
from time import time
import pathlib
from typing import Dict, Union
import glob

import numpy as np
import xarray as xr

from pyschism.mesh.base import Gr3

def combine(var, shape, l2g, name):
    values = np.full(tuple(shape), np.nan)

    local_ids = list(l2g.keys())
    for i, data in enumerate(var):
        cpu_id = str(i).zfill(6) 
        #print(cpu_id)
        n_local_to_global = l2g[cpu_id]
        #print(n_local_to_global[0])
        local_ids = list(n_local_to_global.keys())
        global_idxs = list(
            map(lambda x: int(n_local_to_global[x])-1, local_ids))
        #print(global_idxs[0])
        values[global_idxs] = data
    return values
    

def combine_(dst, var, shape, l2g, name):
    #if len(dst[0][var].shape) < 4:
    out = []
    for i in range(len(dst)):
        out.append(dst[i][var])
    r = combine(out, shape, l2g, name)
    return (xr.DataArray(r, dims = list(dst[0][var].dims), name = var))

class CombineOutputs:

    def __init__(self, path: Union[str, os.PathLike]):
        self.path = pathlib.Path(path)

        if not self.path.exists():
            raise ValueError(f'Directory {self.path} does not exist.')

        nodes = {}
        elements = {}
        for ifile in sorted(self.path.glob(r'local_to_global_[0-9][0-9][0-9][0-9][0-9][0-9]')):
            with open(ifile) as f:
                ns_global, ne_global, np_global, nvrt, nproc, ntracers, \
                    T, S, GEN, AGE, SED3D, EcoSim, ICM, CoSINE, Feco, \
                    TIMOR, FARM, DVD = f.readline().split()
                f.readline()
                # elements
                ne_local = int(f.readline())
                e_local_to_global = {}
                for i in range(ne_local):
                    local_element_id, global_element_id = f.readline().split()
                    e_local_to_global[local_element_id] = global_element_id
                # points
                np_local = int(f.readline())
                n_local_to_global = {}
                for i in range(np_local):
                    local_node_id, global_node_id = f.readline().split()
                    n_local_to_global[local_node_id] = global_node_id
                # sides
                ns_local = int(f.readline())
                s_local_to_global = {}
                for i in range(ns_local):
                    local_side_id, global_side_id = f.readline().split()
                    s_local_to_global[local_side_id] = global_side_id
                f.readline()  # Header:
                line = f.readline().split()
                #old schism print to multiple lines not just one line
                if len(line) != 5:
                    line.extend(f.readline().split())
                self.start_year, self.start_month, self.start_day, \
                    self.start_hour, self.utc_start = line
                nrec, dtout, nspool, nvrt, kz, h0, h_s, h_c, theta_b, \
                    theta_f, ics = f.readline().split()
                #In the old version of schism, ztot was written to nvrt lines
                for i in np.arange(int(nvrt)):
                    f.readline()  # (ztot(k),k=1,kz-1),(sigma(k),k=1,nvrt-kz+1)
                f.readline()  # (ztot(k),k=1,kz-1),(sigma(k),k=1,nvrt-kz+1)
                
                #_ne_local = None
                #_np_local = None
                #while _ne_local != ne_local and _np_local != np_local:
                #    line = f.readline().split()
                #    _np_local = int(float(line[0])
                #    _ne_local = int(float(line[1])
                for i in range(np_local):
                    x, y, z, flag = map(float, f.readline().split())
                    nodes.setdefault(
                        n_local_to_global[str(i+1)], ((x, y), -z))
                for i in range(ne_local):
                    eids = f.readline().split()[1:]
                    elements.setdefault(
                        e_local_to_global[str(i+1)],
                        list(map(lambda x: n_local_to_global[x], eids)))
                nproc_id = ifile.name.split('local_to_global_')[-1]
                self.e_local_to_global.setdefault(nproc_id, e_local_to_global)
                self.n_local_to_global.setdefault(nproc_id, n_local_to_global)
                self.s_local_to_global.setdefault(nproc_id, s_local_to_global)
            #nodes = {str(i+1): nodes[str(i+1)] for i in range(len(nodes))}
            #elements = {str(i+1): elements[str(i+1)] for i in range(len(elements))}
        self.hgrid = Gr3(nodes=nodes, elements=elements, crs='epsg:4326')

    def hotstart(self, it=None):
        self.filenames = sorted(
            self.path.glob(r'hotstart_*_{}.nc'.format(it)))
        dst = []
        for i in range(len(self.filenames)):
            dst.append(xr.open_dataset(self.filenames[i]))
    
        #create dataset
        side = []
        node = []
        elem = []
        one = []
        #variables = ['eta2', 'su2', 'sv2']
        #variables = ['eta2', 'we', 'su2', 'tr_el', 'time', 'it', 'ifile', 'nsteps_from_cold']
        #for var in variables:
        for var in dst[0].variables:
            t0 = time()
            shape = []
            if 'nResident_elem' in dst[0][var].dims:
                shape.append(self.hgrid.elements.array.shape[0])
                if len(dst[0][var].shape) > 1:
                    for i in range(len(dst[0][var].shape)):
                        if i == 0:
                            continue
                        else:
                            shape.append(dst[0][var].shape[i])
                r = combine_(dst, var, shape, self.e_local_to_global, 'nResident_elem')
                elem.append(r)
            elif 'nResident_node' in dst[0][var].dims:
                shape.append(self.hgrid.nodes.values.shape[0])
                if len(dst[0][var].shape) > 1:
                    for i in range(len(dst[0][var].shape)):
                        if i == 0:
                            continue
                        else:
                            shape.append(dst[0][var].shape[i])
                r = combine_(dst, var, shape, self.n_local_to_global, 'nResident_node')
                node.append(r)
            elif 'nResident_side' in dst[0][var].dims:
                shape.append(self.hgrid.elements.sides.shape[0])
                if len(dst[0][var].shape) > 1:
                    for i in range(len(dst[0][var].shape)):
                        if i == 0:
                            continue
                        else:
                            shape.append(dst[0][var].shape[i])
                r = combine_(dst, var, shape, self.s_local_to_global, 'nResident_side')
                side.append(r)
            else:
                one.append(dst[0][var])
            print(f'It took {time()-t0} seconds to combine var {var} in file[{i}]')

        side = xr.merge(side).rename({'nResident_side': 'side'})
        elem = xr.merge(elem).rename({'nResident_elem': 'elem'})
        node = xr.merge(node).rename({'nResident_node': 'node'})
        one = xr.merge(one).rename({'one': 'one_new', 'it': 'iths'})

        xdat = xr.merge([side, elem, node, one])
        #xdat = xr.merge([node, one])
        hfile = 'hotstart_it={}.nc'.format(it)
        xdat.to_netcdf(f'./{hfile}')


    @property
    def n_local_to_global(self):
        if not hasattr(self, '_n_local_to_global'):
            self._n_local_to_global = {}
        return self._n_local_to_global

    @property
    def s_local_to_global(self):
        if not hasattr(self, '_s_local_to_global'):
            self._s_local_to_global = {}
        return self._s_local_to_global

    @property
    def e_local_to_global(self):
        if not hasattr(self, '_e_local_to_global'):
            self._e_local_to_global = {}
        return self._e_local_to_global



import argparse
import logging

from pyschism.cmd import common
from pyschism.mesh.prop import Fluxflag

logger = logging.getLogger(__name__)


class FluxflagCli:

    def __init__(self, args: argparse.Namespace):

        if args.prop_table is not None:
            fluxflag = Fluxflag.from_prop_table(args.hgrid, args.prop_table)

        if args.plot is True:
            fluxflag.make_plot(show=True)

        if args.output_path is not None:
            fluxflag.write(args.output_path, overwrite=args.overwrite)

    @staticmethod
    def add_subparser_action(subparsers):
        add_prop_options_to_parser(subparsers.add_parser('fluxflag'))


def add_prop_options_to_parser(parser):
    common.add_hgrid_to_parser(parser)
    common.add_log_level_to_parser(parser)
    parser.add_argument('--output-path', '-o')
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument('--plot', action='store_true')

    options = parser.add_mutually_exclusive_group(required=True)
    options.add_argument('--prop-table')
    # options.add_argument('--prop-table')
    # table.add_argument('--regtable', dest='table')
    # files = parser.add_mutually_exclusive_group(required=True)
    # files.add_argument(
    #     '--regfiles',
    #     '-r',
    #     help='Generate default tvdflag.prop file.'
    # )
    # files.add_argument(
    #     '--geometry',
    #     nargs='+',
    # )
    # options.add_argument(
    #     '--regfiles',
    #     '--reg',
    #     '-r',
    #     nargs='+',
    #     help='Uses ACE/gredit reg files.'
    # )
    # options.add_argument(
    #     '--propfile',
    #     '--prop',
    #     '-p',
    #     help='Loads a SCHISM prop file.'
    # )
    # options.add_argument(
    #     '--geometry',
    #     '--geom',
    #     '--gpd',
    #     '--json',
    #     '-g',
    #     nargs='+',
    #     help='Any file or files readable by geopandas. Must contain polygons '
    #          'or multipolygon geometries.'
    # )



#! /usr/bin/env python
import argparse
from datetime import datetime
import inspect
import logging
import sys

from pytz import timezone

from pyschism.cmd import *


def init_logger():
    tmp_parser = argparse.ArgumentParser(add_help=False)
    parser_common.add_log_level_to_parser(tmp_parser)
    tmp_args, _ = tmp_parser.parse_known_args()
    if tmp_args.log_level is not None:
        logging.basicConfig(
            format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
            force=True,
        )

        logging.getLogger("pyschism").setLevel({
                "warning": logging.WARNING,
                "info": logging.INFO,
                "debug": logging.DEBUG,
                "critical": logging.CRITICAL,
                "notset": logging.NOTSET,
            }[tmp_args.log_level])
        logging.Formatter.converter = lambda *args: datetime.now(
            tz=timezone("UTC")
        ).timetuple()

        logging.captureWarnings(True)


def main():

    init_logger()
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="clitype")

    clitypes = [
        clitype
        for cliname, clitype in inspect.getmembers(
            sys.modules[__name__], inspect.isclass
        )
        if "Cli" in cliname
    ]

    for clitype in clitypes:
        clitype.add_subparser_action(subparsers)

    args = parser.parse_args()

    for clitype in clitypes:
        if args.clitype.replace("_", "") == f"{clitype.__name__.lower()[:-3]}":
            clitype(args)
            break


if __name__ == "__main__":
    main()



from time import time
from datetime import datetime
import pathlib

from pyschism.mesh.hgrid import Hgrid
from pyschism.forcing.nws.nws2.hrrr2 import AWSZarrInventory

from pyschism.dates import nearest_cycle

'''
Please use it with caution, some variables may have missing data, see:
https://mesowest.utah.edu/html/hrrr/zarr_documentation/html/zarr_variables.html
'''

if __name__ == "__main__":
    t0 = time()
    startdate = datetime(2023, 10, 3)

    #bbox from hgrid
    #hgrid = Hgrid.open('./hgrid.gr3', crs='epsg:4326')
    #bbox = hgrid.bbox
    #or
    ##bbox from list [xmin, ymin, xmax, ymax]
    bbox = [-98, 7.8, -52.986, 42.09]

    outdir = './'
    
    hrrr = AWSZarrInventory(bbox=bbox)
    hrrr.write(outdir=outdir, start_date=startdate, rnday=2, air=True, prc=True, rad=False)
    print(f'It took {(time() - t0)/60} mins')




import argparse
from datetime import timedelta
from enum import Enum
import logging
import pathlib
from pyschism.server.slurm import SlurmConfig

from psutil import cpu_count

from pyschism import dates
from pyschism.cmd import common
from pyschism.driver import ModelConfig
from pyschism.mesh import gridgr3, prop
from pyschism.param.schout import SurfaceOutputVars
from pyschism.server import ServerConfig, SlurmConfig

logger = logging.getLogger(__name__)

CONFIG_FILE_NAME = "config.json"
STATIC_DIRECTORY = "static"


class GridGr3Type(Enum):
    ALBEDO = gridgr3.Albedo
    DIFFMAX = gridgr3.Diffmax
    DIFFMIN = gridgr3.Diffmin
    WATERTYPE = gridgr3.Watertype
    SHAPIRO = gridgr3.Shapiro
    WINDROT = gridgr3.Windrot
    ELEV_IC = gridgr3.ElevIc


class PropType(Enum):
    FLUXFLAG = prop.Fluxflag
    TVDFLAG = prop.Tvdflag


class GridGr3Descriptor:
    def __init__(self, gridgr3_type):
        self.type = gridgr3_type

    def __set__(self, obj, val):
        if val is not None:
            assert isinstance(val, self.type)
        self.gridgr3 = val

    def __get__(self, obj, val):
        if not hasattr(self, "gridgr3"):
            if obj.args.vgrid.is3D():
                return self.type.default(obj.args.hgrid)
        else:
            return self.gridgr3


class ForecastCliMeta(type):
    def __new__(meta, name, bases, attrs):
        attrs["surface_output_vars"] = SurfaceOutputVars()
        for gr3type in GridGr3Type:
            attrs[gr3type.name.lower()] = GridGr3Descriptor(gr3type.value)

        return type(name, bases, attrs)


class ForecastCli(metaclass=ForecastCliMeta):

    start_date = dates.StartDate()
    end_date = dates.EndDate()

    def __init__(self, args: argparse.Namespace):
        self.start_date = dates.nearest_cycle()
        self.args = args
        if self.args.skip_run is True:
            if self.coldstart is not None:
                if self.args.spinup_days is not None:
                    self.coldstart.write(
                        self.coldstart_directory,
                        overwrite=self.args.overwrite,
                    )
                else:
                    self.coldstart.write(
                        self.hotstart_directory, overwrite=self.args.overwrite
                    )

        else:
            if self.coldstart is not None:
                if self.args.spinup_days is not None:
                    self.coldstart.run(
                        self.coldstart_directory, overwrite=self.args.overwrite
                    )
                else:
                    self.coldstart.run(
                        self.hotstart_directory, overwrite=self.args.overwrite
                    )

        if self.args.skip_run is True:
            self.hotstart.write(self.hotstart_directory, overwrite=self.args.overwrite)
        else:
            self.hotstart.run(self.hotstart_directory, overwrite=self.args.overwrite)

    @property
    def coldstart(self):
        if not hasattr(self, "_coldstart"):
            if self.args.vgrid.is2D() is True:
                if self.args.spinup_days is not None:
                    self._coldstart = self.config.coldstart(
                        timestep=self.args.timestep,
                        start_date=self.start_date - self.args.spinup_days,
                        end_date=self.start_date,
                        dramp=self.args.spinup_days,
                        drampbc=self.args.spinup_days,
                        dramp_ss=self.args.spinup_days,
                        drampwafo=self.args.spinup_days,
                        drampwind=self.args.spinup_days,
                        elev_ic=self.args.elev_ic,
                        temp_ic=self.args.temp_ic,
                        salt_ic=self.args.salt_ic,
                        nspool=None,
                        ihfskip=None,
                        nhot_write=None,
                        stations=None,
                        server_config=self.server_config,
                        param_template=self.args.use_param_template,
                        # **self.user_requested_surface_outputs,
                    )
                else:
                    self._coldstart = self.config.coldstart(
                        timestep=self.args.timestep,
                        start_date=self.start_date,
                        end_date=self.args.run_days,
                        elev_ic=self.args.elev_ic,
                        temp_ic=self.args.temp_ic,
                        salt_ic=self.args.salt_ic,
                        nspool=None,
                        ihfskip=None,
                        nhot_write=None,
                        stations=None,
                        server_config=self.server_config,
                        param_template=self.args.use_param_template,
                        # **self.user_requested_surface_outputs,
                    )

            else:
                raise NotImplementedError("Model is 3D.")
        return self._coldstart

    @property
    def hotstart(self):
        return self.config.hotstart(
            self.coldstart,
            timestep=self.args.timestep,
            end_date=self.args.run_days,
            nspool=None,
            ihfskip=None,
            nhot_write=None,
            stations=None,
            server_config=self.server_config,
            param_template=self.args.use_param_template,
            # **self.user_requested_surface_outputs,
        )

    @property
    def config(self):
        if not hasattr(self, "_config"):
            self._config = ModelConfig(
                self.args.hgrid,
                vgrid=self.args.vgrid,
                fgrid=self.args.fgrid,
                albedo=self.args.albedo,
                diffmin=self.args.diffmin,
                diffmax=self.args.diffmax,
                watertype=self.args.watertype,
                fluxflag=self.args.fluxflag,
                tvdflag=self.args.tvdflag,
                iettype=self.args.iettype,
                ifltype=self.args.ifltype,
                itetype=self.args.itetype,
                isatype=self.args.isatype,
                nws=self.args.nws,
                source_sink=self.args.source_sink,
                # waves=self.args.waves,
            )
        return self._config

    @property
    def forecasts_directory(self):
        return self.args.project_directory / "runs"

    @property
    def coldstart_directory(self):
        return self.hotstart_directory / "coldstart"

    @property
    def hotstart_directory(self):
        if not hasattr(self, "_hotstart_directory"):
            timestamp = str(self.start_date).replace(" ", "T")
            self._hotstart_directory = self.forecasts_directory / f"{timestamp}"
            self._hotstart_directory.parent.mkdir(exist_ok=True)
            self._hotstart_directory.mkdir(exist_ok=True)
        return self._hotstart_directory

    @property
    def server_config(self):
        if not hasattr(self, "_server_config"):
            if self.args.workload_manager is None:
                self._server_config = ServerConfig(
                    nproc=self.args.ntasks,
                    # symlink_outputs: str = None,
                    schism_binary=self.args.schism_binary,
                    mpi_launcher=None,
                    modules=self.args.modules,
                    modulepath=self.args.modulepath,
                    modules_init=self.args.modules_init,
                )
            elif self.args.workload_manager == "slurm":
                self._server_config = SlurmConfig(
                    account=self.args.account,
                    ntasks=self.args.ntasks,
                    partition=self.args.partition,
                    walltime=self.args.walltime,
                    filename=self.args.filename,
                    run_directory=None,
                    run_name=None,
                    mail_type=None,
                    mail_user=None,
                    log_filename=None,
                    modules=self.args.modules,
                    modulepath=self.args.modulepath,
                    modules_init=self.args.modules_init,
                    schism_binary=self.args.schism_binary,
                    extra_commands=None,
                    launcher=self.args.launcher,
                    nodes=None,
                    # symlink_outputs=None,
                )
        return self._server_config

    @staticmethod
    def add_subparser_action(subparsers):
        add_forecast_options_to_parser(subparsers.add_parser("forecast"))


def add_forecast_options_to_parser(parser):
    actions = parser.add_subparsers(dest="action")
    add_forecast_init_to_parser(
        actions.add_parser(
            "init",
            help="Initializes a directory for a sequential SCHISM forecast model "
            "deployment.",
        )
    )
    add_forecast_update_to_parser(
        actions.add_parser(
            "update",
            help="Updates a directory  that has been previously initialized for "
            "SCHISM forecast model deployment.",
        )
    )


class ProjectDirectoryAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        tmp_parser = argparse.ArgumentParser(add_help=False)
        tmp_parser.add_argument("--overwrite", action="store_true")
        tmp_args = tmp_parser.parse_known_args()[0]
        project_directory = pathlib.Path(values)
        project_directory.mkdir(exist_ok=tmp_args.overwrite)
        setattr(namespace, self.dest, project_directory)


def add_forecast_init_to_parser(parser):
    parser.add_argument(
        "project_directory",
        action=ProjectDirectoryAction,
        help="System path where the project will be staged.",
    )
    common.add_hgrid_to_parser(parser)
    common.add_vgrid_to_parser(parser)
    common.add_fgrid_to_parser(parser)
    common.add_stratification_to_parser(parser)
    parser.add_argument("--timestep", type=float, required=True)
    parser.add_argument(
        "--forecast-days",
        type=lambda x: timedelta(days=float(x)),
        required=True,
        dest="run_days",
    )
    parser.add_argument(
        "--spinup-days",
        help="Number of days used for model initialization. "
        "Defaults to 15 days spinup.",
        type=lambda x: timedelta(days=float(x)),
    )
    parser.add_argument(
        "--skip-run", action="store_true", help="Skips running the model."
    )
    parser.add_argument("--nproc", type=int, default=cpu_count(logical=False))
    parser.add_argument(
        "--use-param-template",
        # TODO: We need to optionally take a user-provided template
        action="store_true",
    )
    common.add_gridgr3_to_parser(parser)
    common.add_ic_to_parser(parser)
    common.add_prop_to_parser(parser)
    #common.add_ibctype_to_parser(parser)
    common.add_tidal_constituents_to_parser(parser)
    common.add_tidal_database_to_parser(parser)
    #common.add_baroclinic_database_to_parser(parser)
    common.add_bctides_options_to_parser(parser)
    common.add_nws_to_parser(parser)
    common.add_source_sink_to_parser(parser)
    # common.add_waves_to_parser(parser)
    common.add_surface_outputs_to_parser(parser)
    common.add_stations_outputs_to_parser(parser)
    common.add_log_level_to_parser(parser)
    common.add_schism_binary_to_parser(parser)
    common.add_modules_to_parser(parser)
    common.add_workload_manager_options_to_parser(parser)
    parser.add_argument(
        "--overwrite", action="store_true", help="Allow overwrite of output directory."
    )


def add_forecast_update_to_parser(parser):
    parser.add_argument(
        "project_directory", help="System path of the staged directory to update."
    )


# ------ drafts ----------


#     init.add_argument('hgrid', help='Horizontal grid file.')
#     init.add_argument('--fgrid', help='Friction grid file.')

#     mesh_options = init.add_argument_group('mesh_options')
#     mesh_options.add_argument('--hgrid-crs')
#     mesh_options.add_argument('--fgrid-crs')
#     mesh_options.add_argument(
#         '--fgrid-type',
#         choices=['auto', 'manning', 'drag', 'rough'],
#         default='auto')
#     init.add_argument("--timestep", type=float, required=True)
#     init.add_argument("--forecast-days", type=float, required=True)
#     init.add_argument(
#         "--spinup-days",
#         help="Number of days used for model initialization. "
#         "Defaults to 15 days spinup.",
#         type=lambda x: timedelta(days=float(x)),
#     )
#     add_prop_options_to_parser(init)
#     init.add_argument(
#         "--skip-run", action="store_true",
#         help="Skips running the model.")
#     init.add_argument('--nproc', type=int, default=cpu_count(logical=False))
#     init.add_argument('--use-param-template', action='store_true')
#     init.add_argument(
#         "--log-level",
#         choices=['info', 'warning', 'debug'],
#         default='warning'
#     )
#     init.add_argument(
#         "--overwrite", action="store_true",
#         help="Allow overwrite of output directory.")
#     _add_tidal_constituents(init)
#     _add_atmospheric_forcing(init)
#     _add_hydrologic_forcing(init)
#     init.add_argument('--shapiro', action='store_true')
#     init.add_argument('--shapiro-bin', action='store_true')

#     # TODO: Additional forcings.
#     # _add_wave_forcing(forecast)
#     model_outputs = init.add_argument_group('model_outputs')
#     # TODO: Stations outputs.
#     _add_stations_outputs(model_outputs)
#     _add_surface_outputs(model_outputs)
#     server_config = init.add_subparsers(dest="server_config")
#     slurm = server_config.add_parser(
#         'slurm', help="Add options for slurm run configuration.")
#     slurm.add_argument('--account')
#     slurm.add_argument('--partition')
#     slurm.add_argument('--walltime', type=float, help="In hours, float.")
#     slurm.add_argument('--slurm-filename')
#     slurm.add_argument('--slurm-rundir')
#     slurm.add_argument('--run-name')
#     slurm.add_argument('--mail-type')
#     slurm.add_argument('--mail-user')
#     slurm.add_argument('--log-filename')
#     slurm.add_argument('--path-prefix')
#     slurm.add_argument('--slurm-nodes')
#     slurm.add_argument('--slurm-launcher', default='srun')
#     slurm.add_argument('--extra-commands', action='append')
#     slurm.add_argument(
#         '--module',
#         default=list(),
#         action='append',
#         dest='modules')


# def add_forecast_update(actions):
#     update = actions.add_parser("update")
#     update.add_argument("project_directory")

# def _add_tidal_constituents(parser):
#     tides = parser.add_argument_group('tides')
#     options = tides.add_mutually_exclusive_group()
#     options.required = True
#     options.add_argument("--all-constituents", action="store_true")
#     options.add_argument("--major-constituents", action="store_true")
#     options.add_argument(
#         "-c", "--constituents",
#         action='append',
#         # choices=["K1", "O1", "P1", "Q1", "MM", "Mf", "M4", "MN4", "MS4",
#         #         "2N2", "S1"],
#         choices=["Z0", "K2", "S2", "M2", "N2", "K1", "P1", "O1", "Q1"],
#         dest='constituents',
#         default=False,
#         help="Tidal constituent to be forced in the model (case-insensitive).")
#     tides.add_argument("--tidal-database", choices=['tpxo', 'hamtide'],
#                        default='tpxo')
#     tides.add_argument("--include-tidal-velocity",
#                        "--bnd-vel", action="store_true", dest='bnd_vel')
#     tides.add_argument("--Z0", type=float)


# def _add_atmospheric_forcing(parser):
#     atmospheric_forcing_1 = parser.add_argument_group(
#         'atmospheric forcing level 1')
#     data_source_1 = atmospheric_forcing_1.add_mutually_exclusive_group()
#     for product in ForecastProduct:
#         data_source_1.add_argument(
#             f"--{product.value.lower().replace('_', '-')}",
#             action="store_true",
#             help=f'Use {product.value} as data source.')
#     data_source_1.add_argument(
#         "--gfs", action="store_true",
#         help="Alias for --gfs-0p25-1hr")
#     data_source_1.add_argument(
#         "--gdas", action='store_true',
#         help="Alias for --gdas-0p25",
#     )
#     atmospheric_forcing_1.add_argument(
#         "--no-air-1", action="store_true",
#         help="Disables air mass forcing option.")
#     atmospheric_forcing_1.add_argument(
#         "--no-prc-1", action='store_true',
#         help="Disables precipitation option.")
#     atmospheric_forcing_1.add_argument(
#         "--no-rad-1", action='store_true',
#         help="Disables solar radiation flux option.")

#     atmospheric_forcing_2 = parser.add_argument_group(
#         'atmospheric forcing level 2')
#     data_source_2 = atmospheric_forcing_2.add_mutually_exclusive_group()
#     for product in ["hrrr"]:
#         data_source_2.add_argument(
#             f"--{product}",
#             help=f'Use {product.upper()} as data source.',
#             action='store_true'
#         )
#     atmospheric_forcing_2.add_argument(
#         "--no-air-2", action="store_true",
#         help="Disables air mass forcing option.")
#     atmospheric_forcing_2.add_argument(
#         "--no-prc-2", action='store_true',
#         help="Disables precipitation option.")
#     atmospheric_forcing_2.add_argument(
#         "--no-rad-2", action='store_true',
#         help="Disables solar radiation flux option.")


# def _add_hydrologic_forcing(parser):
#     src_snk = parser.add_argument_group('Sources and sinks')
#     src_snk.add_argument("--hydrology", action="append",
#                          help="Add source and sink flow.",
#                          choices=['NWM'],
#                          default=[])


# def _add_wave_forcing(parser):
#     wave_forcing = parser.add_argument_group('Waves')
#     data_source = wave_forcing.add_mutually_exclusive_group()
#     for product in ["WWIII"]:
#         data_source.add_argument(f'--{product}')



from argparse import Namespace
import sys

from pyschism.cmd import common
from pyschism.outputs.outputs import OutputsCollector


class OutputsCli:

    def __init__(self, args: Namespace):
        outputs = OutputsCollector(args.outputs_directory)
        if args.action == 'plot':
            getattr(outputs, args.variable_name).make_plot(show=True)

        elif args.action == 'animate':
            getattr(outputs, args.variable_name).animation(
                vmin=args.vmin,
                vmax=args.vmax,
                show=args.show,
                fps=args.fps,
                save=args.save,
            )

    @staticmethod
    def add_subparser_action(subparsers):
        add_outputs_options_to_parser(subparsers.add_parser('outputs'))


def add_outputs_options_to_parser(parser):

    parser.add_argument('outputs_directory')
    parser.add_argument('variable_name')
    subparsers = parser.add_subparsers(dest='action')
    add_plot_action(subparsers)
    add_animate_action(subparsers)


def add_plot_action(subparsers):
    parser = subparsers.add_parser('plot')
    parser.add_argument('step')


def add_animate_action(subparsers):
    parser = subparsers.add_parser('animate')
    common.add_vmin_to_parser(parser)
    common.add_vmax_to_parser(parser)
    show = parser.add_mutually_exclusive_group() 
    show.add_argument('--show', dest='show', action='store_true')
    show.add_argument('--no-show', dest='show', action='store_false')
    parser.set_defaults(show=False if "--save" in " ".join(sys.argv) else True)
    parser.add_argument('--fps', type=float)
    parser.add_argument('--save')
    parser.add_argument('--figsize', nargs=2, type=float)



#! /usr/bin/env python
from datetime import datetime, timedelta
import logging
import os
import pathlib
import shutil
import tarfile
import tempfile
import unittest
import urllib.request
import f90nml

from pyschism import dates
from pyschism.mesh import Hgrid, Vgrid
from pyschism.driver import ModelConfig


logging.basicConfig(level=logging.INFO, force=True)


class ModelConfigurationTestCase(unittest.TestCase):

    def test_ncor_zero(self):

        hgrid=Hgrid.open('https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll', crs='epsg:26918')

        config = ModelConfig(
            hgrid,
            flags = [[3, 3, 0, 0]],
        )
        
        # create reference dates
        spinup_time = timedelta(days=0.)
        start_date=datetime(2023,9,10)

        # create a coldstart object
        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + timedelta(days=3.),
            timestep=150.,
            dramp=spinup_time,
            dramp_ss=spinup_time,
            drampwind=spinup_time,
            nspool=timedelta(hours=1),
        )

        with tempfile.TemporaryDirectory() as tempdir:
            coldstart.write(tempdir, overwrite=True)
            
            nml = f90nml.read(f'{tempdir}/param.nml')
            assert('ncor' in nml['opt'])
            assert(nml['opt']['ncor'] == 0) 

    def test_ncor_one(self):

        hgrid=Hgrid.open('https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll', crs='epsg:4326')

        config = ModelConfig(
            hgrid,
            flags = [[3, 3, 0, 0]],
        )
        
        # create reference dates
        spinup_time = timedelta(days=0.)
        start_date=datetime(2023,9,10)

        # create a coldstart object
        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + timedelta(days=3.),
            timestep=150.,
            dramp=spinup_time,
            dramp_ss=spinup_time,
            drampwind=spinup_time,
            nspool=timedelta(hours=1),
        )

        with tempfile.TemporaryDirectory() as tempdir:
            coldstart.write(tempdir, overwrite=True)

            nml = f90nml.read(f'{tempdir}/param.nml')
            assert('ncor' in nml['opt'])
            assert(nml['opt']['ncor'] == 1) 

if __name__ == '__main__':
    unittest.main()



from datetime import datetime
import logging

from pyschism.mesh.hgrid import Hgrid
from pyschism.forcing.hycom.hycom2schism import OpenBoundaryInventory

'''
outputs:
    elev2D.th.nc (elev=True)
    SAL_3D.th.nc (TS=True)
    TEM_3D.th.nc (TS=True)
    uv3D.th.nc   (UV=True)
'''

logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logger = logging.getLogger('pyschism')
logger.setLevel(logging.INFO)

if __name__ == "__main__":
    hgrid=Hgrid.open('./hgrid.gr3', crs='epsg:4326')
    vgrid='./vgrid.in'
    outdir='./'
    start_date=datetime(2018, 8, 1)
    rnday=20

    #boundary
    bnd=OpenBoundaryInventory(hgrid, vgrid)

    #ocean_bnd_ids - segment indices, starting from zero.
    bnd.fetch_data(outdir, start_date, rnday, elev2D=True, TS=True, UV=True, ocean_bnd_ids=[0,1,2])



from datetime import datetime, timedelta
from functools import cached_property
import pathlib
import logging
from typing import Union

from pyschism.forcing.bctides import Tides

logger = logging.getLogger(__name__)

class Bctides:
    """Bctides class
    This is the bctides class to generate bctides.in file.
    
    """

    def __init__(
        self,
        hgrid,
        flags: list = None,
        constituents: Union[str, list] = 'major',
        database: str = 'tpxo',
        add_earth_tidal: bool = True,
        cutoff_depth: float = 50.0,
        ethconst: list = None,
        vthconst: list = None,
        tthconst: list = None,
        sthconst: list = None,
        tobc: list = None,
        sobc: list = None,
        relax: list = None,
    ):
        """Initialize Bctides ojbect
        Parameters
        --------
        hgrid: Hgrid object
        flags: nested list of bctypes
        constituents: str or list
        database: str ('tpxo' or 'fes2014')
        add_earth_tidal: bool
        cutoff_depth: float
        ethconst: list (constant elevation value for each open boundary)
        vthconst: list (constant discharge value for each open boundary)
        tthconst: list (constant temperature value for each open boundary)
        sthconst: list (constant salinity value for each open boundary)
        tobc: list (nuding factor of temperature for each open boundary)
        sobc: list (nuding factor of salinity for each open boundary)
        realx: list (relaxation constants for inflow and outflow)
        """

        self.hgrid = hgrid
        self.add_earth_tidal = add_earth_tidal
        self.cutoff_depth = cutoff_depth
        self.tides = Tides(constituents=constituents, tidal_database=database)
        self.flags = flags
        self.ethconst = ethconst 
        self.vthconst = vthconst 
        self.tthconst = tthconst 
        self.sthconst = sthconst 
        self.tobc = tobc
        self.sobc = sobc
        self.relax = relax

    def __str__(self):
 
        #first line in the bctides.in is a note, not used in the schism code
        f = [
            f"!{str(self.start_date)} UTC",
        ]

        #get earth tidal potential and frequency
        if self.add_earth_tidal:
            f.append(f"{self.ntip} {self.cutoff_depth} !number of earth tidal potential, cut-off depth for applying tidal potential")
            for constituent in self.tides.get_active_potential_constituents(): 
                forcing = self.tides(self.start_date, self.rnday, constituent)
                f.append(
                    " ".join([
                        f"{constituent}\n",
                        f"{forcing[0]:G}",
                        f"{forcing[1]:G}",
                        f"{forcing[2]:G}",
                        f"{forcing[3]:G}",
                        f"{forcing[4]:G}",
                    ])
                )
        else:
            f.append(f"0 {self.cutoff_depth} !number of earth tidal potential, cut-off depth for applying tidal potential")

        #get tidal boundary 
        f.append(f"{self.nbfr:d} !nbfr")
        if self.nbfr > 0:
            for constituent in self.tides.get_active_forcing_constituents():
                forcing = self.tides(self.start_date, self.rnday, constituent)
                f.append(
                    " ". join([
                        f"{constituent}\n", 
                        f"{forcing[2]:G}",
                        f"{forcing[3]:G}",
                        f"{forcing[4]:G}",

                    ])
                )
 
        #get amplitude and phase for each open boundary         
        f.append(f"{len(self.gdf)} !nope")
        if len(self.gdf) != len(self.flags):
            raise ValueError(f'Number of open boundary {len(self.gdf)} is not consistent with number of given bctypes {len(self.flags)}!')
        for ibnd, (boundary, flag) in enumerate(zip(self.gdf.itertuples(), self.flags)):
            logger.info(f"Processing boundary {ibnd+1}:")
            #number of nodes and flags
            line = [
                f"{len(boundary.indexes)}",
                *[str(digit) for digit in flag],
                f"!open bnd {ibnd+1}",
            ]
            f.append(" ".join(line))

            #It only accounts for elevation, velocity, temperature, salinity
            #TODO: add information for tracers
            if len(flag) > 4:
                raise NotImplementedError(f"Tracer module is not implemented yet!") 
            iettype, ifltype, itetype, isatype = [i for i in flag]

            #elevation boundary
            logger.info(f"Elevation type: {iettype}")
            if iettype == 1:
                logger.warning(f'time history of elevation is read in from elev.th (ASCII)!')
            elif iettype == 2:
                logger.info("You are choosing type 2 for elevation, value is {selfethconst[ibnd]} ")
                f.append(f"{self.ethconst[ibnd]}")
            elif iettype == 4:
                logger.warning('time history of elevation is read in from elev2D.th.nc (netcdf)')
            elif iettype == 3 or iettype == 5:
                if iettype == 5:
                    logger.warning(f'Combination of 3 and 4, time history of elevation is read in from elev2D.th.nc!')
                for constituent in self.tides.get_active_forcing_constituents():
                    f.append(f"{constituent}")
                    vertices = self.hgrid.get_xy(crs=self.hgrid.crs)[boundary.indexes, :]
                    amp, phase = self.tides.get_elevation(constituent, vertices)
                    for i in range(len(boundary.indexes)):
                        f.append(f"{amp[i]: .6f} {phase[i]: .6f}")
            elif iettype == 0:
                logger.warning(f"elevations are not specified for this boundary (in this case the discharge must be specified)")
            else:
                raise IOError(f"Invalid type {iettype} for elevation!")

            #velocity
            logger.info(f"Velocity type: {ifltype}")
            if ifltype == 0:
                logger.info("Velocity is not sepcified, not input needed!")
            elif ifltype == 1:
                logger.warning("time history of discharge is read in from flux.th (ASCII)!")
            elif ifltype == 2:
                logger.info("You are choosing type 2 for velocity, value is {self.vthconst[ibnd]} ")
                f.append(f"{self.vthconst[ibnd]}")
            elif ifltype == 3 or ifltype == 5:
                if ifltype == 5:
                    logger.warning(f'Combination of 3 and 4, time history of velocity is read in from uv.3D.th.nc!')
                for constituent in self.tides.get_active_forcing_constituents():
                    f.append(f"{constituent}")
                    vertices = self.hgrid.get_xy(crs=self.hgrid.crs)[boundary.indexes, :]
                    uamp, uphase, vamp, vphase = self.tides.get_velocity(constituent, vertices)
                    for i in range(len(boundary.indexes)):
                        f.append(f"{uamp[i]: .6f} {uphase[i]: .6f} {vamp[i]: .6f} {vphase[i]: 6f}")
            elif ifltype == 4 or -4:
                logger.warning("time history of velocity (not discharge!) is read in from uv3D.th.nc (netcdf)")
                if ifltype == -4:
                    logger.info(f"You are using type -4, relaxation constants for inflow  is {self.relax[0]}, for outflow is {self.relax[1]}")
                    f.append(f"{self.relax[0]} {self.relax[1]} !relaxation constant")
            elif ifltype == -1: 
                raise NotImplementedError(f"Velocity type {ifltype} not implemented yet!")
                #logger.info(f"Flather type radiation b.c. (iettype must be 0 in this case)!")
                #f.append(['vn_mean'])
                
                #TODO: add mean normal velocity at the node (at all levels)[
            else:
                raise IOError(f"Invalid type {ifltype} for veloicty!")

            #temperature
            logger.info(f"Temperature type: {itetype}")
            if itetype == 0:
                logger.warning("Temperature is not sepcified, not input needed!")
            elif itetype == 1:
                logger.warning("time history of temperature will be read in from TEM_1.th!")
                logger.info(f"Nudging factor for T at boundary {ibnd+1} is {self.tobc[ibnd]}")
                f.append(f"{self.tobc[ibnd]} !nudging factor for T")
            elif itetype == 2:
                logger.info("You are choosing type 2 for temperature, value is {self.tthconst[ibnd]} ")
                f.append(f"{self.tthconst[ibnd]} !T")

                logger.info(f"Nudging factor for T at boundary {ibnd+1} is {self.tobc[ibnd]}")
                f.append(f"{self.tobc[ibnd]} !nudging factor for T")
            elif itetype == 3:
                logger.info("Using initial temperature profile for inflow")
                logger.info(f"Nudging factor for T at boundary {ibnd+1} is{self.tobc[ibnd]}")
                f.append(f"{self.tobc[ibnd]} !nudging factor for T")
            elif itetype == 4:
                logger.warning("time history of temperature is read in from TEM_3D.th.nc (netcdf)!")
                logger.info(f"Nudging factor for T at boundary {ibnd+1} is{self.tobc[ibnd]}")
                f.append(f"{self.tobc[ibnd]} !nudging factor for T")
            else:
                raise IOError(f"Invalid type {itetype} for salinity!")        

            #salinity
            logger.info(f"Salinity type: {isatype}")
            if isatype == 0:
                logger.info("Salinity is not sepcified, not input needed!")
            elif isatype == 1:
                logger.warning("time history of salinity will be read in from SAL_1.th!")
                logger.info(f"Nudging factor for salt at boundary {ibnd+1} is {self.sobc[ibnd]}")
                f.append(f"{self.sobc[ibnd]} !nudging factor for S")
            elif isatype == 2:
                logger.info("Yor are choosing type 2 for salinity, value is {self.sthconst[ibnd]} ")
                f.append(f"{self.sthconst[ibnd]} !S")

                logger.info(f"Nudging factor for salt at boundary {ibnd+1} is {self.sobc[ibnd]}")
                f.append(f"{self.sobc[ibnd]} !nudging factor for S")
            elif isatype == 3:
                logger.info("Using initial salinity profile for inflow")
                logger.info(f"Nudging factor for salt at boundary {ibnd+1} is {self.sobc[ibnd]}")
                f.append(f"{self.sobc[ibnd]} !nudging factor for S")
            elif isatype == 4:
                logger.warning("time history of salinity is read in from SAL_3D.th.nc (netcdf)!")
                logger.info(f"Nudging factor for salt at boundary {ibnd+1} is {self.sobc[ibnd]}")
                f.append(f"{self.sobc[ibnd]} !nudging factor for S")
            else:
                raise IOError(f"Invalid type {isatype} for salinity!") 
                
        return "\n".join(f)

    def write(
        self,
        output_directory,
        start_date: datetime = None,
        rnday = None,
        overwrite: bool = True, 
    ):
        """
        parameters
        --------
        output_directory: str
        start_date: datetime.datetime
        rnday: int, float or datetime.timedelta
        overwrite: bool 
        """

        if start_date is not None:
            self.start_date = start_date
        else:
            raise IOError("Please specify the start_date!")

        if rnday is not None:
            self.rnday = rnday
        else:
            raise IOError("Please specify the number of days!")

        output_directory = pathlib.Path(output_directory)
        output_directory.mkdir(exist_ok=overwrite, parents=True) 

        bctides = output_directory / "bctides.in"

        if bctides.exists() and not overwrite:
            raise IOError("path exists and overwrite is False") 
        with open(bctides, "w") as f:
            f.write(str(self))

    @cached_property
    def gdf(self):
       return self.hgrid.boundaries.open.copy()
        
    @property
    def ntip(self):
        return len(self.tides.get_active_potential_constituents())

    @property
    def nbfr(self):
        return len(self.tides.get_active_potential_constituents())




from argparse import Namespace
import pathlib

from pyschism.mesh import Hgrid


class Grd2SmsCli:

    def __init__(self, args: Namespace):

        if args.output_path.exists() and args.overwrite is not True:
            raise IOError('File exists and overwrite is not True.')

        Hgrid.open(args.hgrid, crs=args.hgrid_crs).write(
            args.output_path, args.overwrite, format='2dm')

    @staticmethod
    def add_subparser_action(subparser):
        add_grd2sms(subparser.add_parser('grd2sms'))


def add_grd2sms(parser):
    parser.add_argument('hgrid', type=pathlib.Path)
    parser.add_argument('output_path', type=pathlib.Path)
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument('--hgrid-crs')
